package SEU.module.algorithm;

import adf.agent.action.common.ActionMove;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.algorithm.PathPlanning;

import rescuecore2.misc.Pair;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import java.awt.Polygon;
import java.util.*;

public class SEUPathPlanning extends PathPlanning {

    private Map<EntityID, Set<EntityID>> graph;
    private EntityID from;
    private Collection<EntityID> targets;
    private List<EntityID> result;
    private List<EntityID> previousPath = new ArrayList<>();
    private EntityID previousTarget = null;
    private int continueMovingTime = 0;
    public Map<Integer, StandardMessageURN> actionHistory = new TreeMap<Integer, StandardMessageURN>();
	public Map<Integer, Pair<Integer, Integer>> locationHistory= new TreeMap<Integer, Pair<Integer, Integer>>();
	protected Map<Integer, EntityID> positionHistory = new TreeMap<Integer, EntityID>();
	protected int time=this.agentInfo.getTime();
	EntityID lastposition=null;
	 protected StandardWorldModel world;
	 private static final boolean DO_NOT_MOVE_IN_BURNING_BUILDING = true;

    public SEUPathPlanning(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        this.init();
    }
    
    public class Node implements Comparable<Node>{  
   	 private double routeCost;
   	 private double g;
   	 private double h;
   	 private Node parent;
   	private EntityID point;
   	
   	public Node(EntityID id) {
   		this.point = id;
        this.setParent(null);
        this.routeCost = 0.0;
		this.g = 0;
		this.h = 0;
	}
   	public EntityID getId() {
        return point;
   	}
       
 	public void setG(double g) {
        this.g = g;
    }
 	public void setH(double d) {
        this.h = d;
    }
 	
   	public void setCost(double cost) {
        this.routeCost = cost;
    }
   	
    public double getG() {
   	    return g ;
       }
    public double getH() {
   	    return h ;
       }
   	 public double getcost() {
       	    return routeCost ;
           }
    
   	 public void setParent(Node parent) {
   		    this.parent = parent;
   	    }
   	
   	 public Node getParent() {
   			return parent;
   		}

		@Override
		public int compareTo(Node route) {
			return (int) (this.getcost() - route.getcost());
		} 
   	}

    private void init() {
        Map<EntityID, Set<EntityID>> neighbours = new LazyMap<EntityID, Set<EntityID>>() {
            @Override
            public Set<EntityID> createValue() {
                return new HashSet<>();
            }
        };
        for (Entity next : this.worldInfo) {
            if (next instanceof Area) {
                Collection<EntityID> areaNeighbours = ((Area) next).getNeighbours();
                neighbours.get(next.getID()).addAll(areaNeighbours);
            }
        }
        this.graph = neighbours;
    }

    @Override
    public List<EntityID> getResult() {
        return this.result;
    }

    @Override
    public PathPlanning setFrom(EntityID id) {
        this.from = id;
        return this;
    }

    @Override
    public PathPlanning setDestination(Collection<EntityID> targets) {
        this.targets = targets;
        return this;
    }

    @Override
    public PathPlanning updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        return this;
    }

    @Override
    public PathPlanning precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        return this;
    }

    @Override
    public PathPlanning resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        return this;
    }

    @Override
    public PathPlanning preparate() {
        super.preparate();
        return this;
    }
 
    @Override
    public PathPlanning calc() {
        result = null;
        if (this.from == null ) {       	
        	return null;
        }
			if (targets == null)  {
	        	return null;
	        }
		if (targets.isEmpty()){
        	return null;
        }
        List<EntityID> path;     
        double distance=Double.MAX_VALUE;
        EntityID finallytarget=null;
        for (EntityID target: targets) {
            if(distance>this.worldInfo.getDistance(this.from, target)){
        		distance=this.worldInfo.getDistance(this.from, target);
            	finallytarget=target;
        	}
        	
        }
             // path=planMove(this.from, finallytarget);
            path = getShortestPath(this.from, finallytarget);
            if (path != null && !path.isEmpty()) {         
                result = path;               
            }else{           	
                System.out.println("ERROR: " + agentInfo.toString() + " path is null......");	
            }
        return this;
    }
    
    public List<EntityID> getShortestPath(EntityID source, EntityID destination) {
        if (destination == null) {
            System.err.println("ERROR: " + agentInfo.toString() + " Destination is null......");
            return new ArrayList<EntityID>();
        }        
        Node sourceNode=new Node(source);
        Node destinationNode=new Node(destination);
        Set<Node> open = new HashSet<Node>();
        Set<EntityID> closed = new HashSet<EntityID>();
        Node current;
        sourceNode.setG(0);
        sourceNode.setCost(0);
        sourceNode.setParent(null);
        destinationNode.setParent(null);
        open.add(sourceNode);
        boolean found = false;
        if (sourceNode.getId().equals(destinationNode.getId())) {           
            found = true;          
            return getPath(sourceNode);
        } 
        Human me=(Human)this.agentInfo.me();
        if(me instanceof PoliceForce){
        while (open.size()!= 0) {
            current = Collections.min(open);
            if (current.getId().equals(destinationNode.getId())) {
            	destinationNode.setParent(current.getParent());
            	found = true;
                break;
            } 
            open.remove(current);
            closed.add(current.getId());
            Collection<EntityID> neighbours = graph.get(current.getId());
            if (neighbours.isEmpty()) {
                continue;
            }
            for (EntityID neighbour : neighbours) {
            	 Node neighbourNode=new Node(neighbour);
            	 
            	if (!closed.contains(neighbour)) {
                double neighbourG = current.getG()+this.worldInfo.getDistance(neighbour, current.getId()); // neighbour weight
                
                if (DO_NOT_MOVE_IN_BURNING_BUILDING) {
                    Area area = (Area) worldInfo.getEntity(neighbour);
                    if (area != null && (area instanceof Building) && ((Building) area).isFierynessDefined()) {
                        int fieriness = ((Building) area).getFieryness();
                        if (fieriness > 0 && fieriness < 4) {
                            neighbourG*= 10000;
                        }
                    }
                }
                if (!iscon(neighbourNode,open)){
                	neighbourNode.setParent(current);
                	neighbourNode.setH(this.worldInfo.getDistance(neighbour, destination));
                	neighbourNode.setG(neighbourG);
                	neighbourNode.setCost(neighbourNode.getH() + neighbourG);
                	open.add(neighbourNode);                	
                }
                else{
                    if (neighbourNode.getG() > neighbourG) {
                        neighbourNode.setParent(current);
                        neighbourNode.setG(neighbourG);  
                        neighbourNode.setCost(neighbourNode.getH() + neighbourG);                      
                    }                             
                }
            	}
            }
           
        }}else{
        	 while (open.size()!= 0) {
                 current = Collections.min(open);
                 if (current.getId().equals(destinationNode.getId())) {
                 	destinationNode.setParent(current.getParent());
                 	found = true;
                     break;
                 } 
                 open.remove(current);
                 closed.add(current.getId());
                 Collection<EntityID> neighbours = graph.get(current.getId());
                 if (neighbours.isEmpty()) {
                     continue;
                 }
                 for (EntityID neighbour : neighbours) {
                 	 Node neighbourNode=new Node(neighbour);
                 	 
                 	if (!closed.contains(neighbour)) {
                     double neighbourG = current.getG()+this.worldInfo.getDistance(neighbour, current.getId()); // neighbour weight
                     if(isunpassable(current.getId(),neighbour)){
                    	 neighbourG=current.getG()+1000000*(this.worldInfo.getDistance(neighbour, current.getId())+1000);
                     }
                     if (DO_NOT_MOVE_IN_BURNING_BUILDING) {
                         Area area = (Area) worldInfo.getEntity(neighbour);
                         if (area != null && (area instanceof Building) && ((Building) area).isFierynessDefined()) {
                             int fieriness = ((Building) area).getFieryness();
                             if (fieriness > 0 && fieriness < 4) {
                                 neighbourG *= 10000;
                             }
                         }
                     }
                     if (!iscon(neighbourNode,open)){
                     	neighbourNode.setParent(current);
                     	neighbourNode.setH(this.worldInfo.getDistance(neighbour, destination));
                     	neighbourNode.setG(neighbourG);
                     	neighbourNode.setCost(neighbourNode.getH() + neighbourG);
                     	open.add(neighbourNode);                	
                     }
                     else{
                         if (neighbourNode.getG() > neighbourG) {
                             neighbourNode.setParent(current);
                             neighbourNode.setG(neighbourG);  
                             neighbourNode.setCost(neighbourNode.getH() + neighbourG);                      
                         }                             
                     }
                 	}
                 }
                
             }
        }
        if (found) {
            return getPath(destinationNode);
        } else {
            return new ArrayList<EntityID>();
        }
    }
    
    public boolean iscon(Node node , Set<Node> open){
    	if(open.isEmpty())
    		return false;
    	for(Node node1:open){
    		if(node.getId()==node1.getId())
    			return true;
    	}
    	return false;
    }

        private List<EntityID> getPath(Node destinationNode) {
          
            List<EntityID> path = new LinkedList<EntityID>();            
            Node current=destinationNode;
            path.add(current.getId());
            while (current.getParent() != null) {
                path.add(0,current.getParent().getId());
                current = current.getParent();
            }           
            return path;
        }

        public boolean isBlocked() {
		if (time< 4)
			return false;
		if(this.lastposition==null)
			return false;
		if (this.agentInfo.getExecutedAction(time-1).getClass() == ActionMove.class &&
			this.agentInfo.getExecutedAction(time-2).getClass() == ActionMove.class) {
			Pair<Integer, Integer> location_1 = locationHistory.get(time - 2);
			Pair<Integer, Integer> location_2 = locationHistory.get(time - 1);
			Pair<Integer, Integer> location_3 = locationHistory.get(time);
			
			if (location_1 == null || location_2 == null || location_3 == null)
				return false;
			
			double distance_1 = getDistance(location_1, location_2);
			double distance_2 = getDistance(location_2, location_3);
			
			if (distance_1 < 1000 && distance_2 < 1000) {
				return true;
			}
		}
		
		return false;
	}

    public static double getDistance(Pair<Integer, Integer> pair1, Pair<Integer, Integer> pair2) {
		int x1 = pair1.first().intValue();
		int y1 = pair1.second().intValue();
		int x2 = pair2.first().intValue();
		int y2 = pair2.second().intValue();
		return Math.hypot((x2 - x1) * 1.0, (y2 - y1) * 1.0);
}

    private boolean isGoal(EntityID e, Collection<EntityID> test) {
        return test.contains(e);
    }

    public boolean isunpassable(EntityID from,EntityID to){
    	Area ar=(Area)(this.worldInfo.getEntity(to));
    	if(!ar.isBlockadesDefined())
    	return false;
    	List<EntityID>blockades=ar.getBlockades();
    	if(blockades==null)
    		return false;
    	if(blockades.isEmpty())
    		return false;
    	List<Blockade> blockedStart = new ArrayList<Blockade>();
        List<Blockade> blockedEnd = new ArrayList<Blockade>();
        List<Blockade> blockededge1 = new ArrayList<Blockade>();
        List<Blockade> blockededge2 = new ArrayList<Blockade>();
    	  for (int i = 0; i < ar.getEdges().size() - 1; i++) {    		  
              Edge edge1 = ar.getEdges().get(i);
              if(edge1.getNeighbour()==from){
              if(edge1.getNeighbour()!=null){
            	  for (EntityID blockade : blockades) {
            		  Blockade blockade1=(Blockade)this.worldInfo.getEntity(blockade);
                      if (distance((Polygon)blockade1.getShape(), edge1.getStart()) < 750) {
                          blockedStart.add(blockade1);
                      }
                      if (distance((Polygon)blockade1.getShape(), edge1.getEnd()) <750) {
                          blockedEnd.add(blockade1);
                      }
                  }
            	  if (blockades.size() == 1) {
                      if (containsEach(blockedEnd, blockedStart)) 
                    	  return true;   
                  } else {
                      for (Blockade block1 : blockedStart) {
                          for (Blockade block2 : blockedEnd) {                           
                              if (isblockPassable((Polygon)block1.getShape(), (Polygon)block2.getShape(), 750 ))
                            	  return true;   
                          }
                      }
                  }
            	 
              }
              }
    	  }
    	  if(ar.getEdges().size()==4 && ar instanceof Road) {
              for (int i = 0; i < ar.getEdges().size() - 1; i++) {
                  Edge edge1 = ar.getEdges().get(i);
                  if (edge1.getNeighbour() == from) {
                      for (int j = 0; j < ar.getEdges().size() - 1; j++) {
                          for (int m = 0; m < ar.getEdges().size() - 1; m++) {
                              Edge edge2 = ar.getEdges().get(j);
                              Edge edge3 = ar.getEdges().get(m);
                              if(edge2!=edge1 && edge3!=edge1 && (edge1.getStart()==edge2.getStart() ||edge1.getStart()==edge2.getEnd()) && (edge1.getEnd()==edge2.getStart() ||edge1.getEnd()==edge2.getEnd())){
                                  for (EntityID blockade : blockades) {
                                      Blockade blockade1=(Blockade)this.worldInfo.getEntity(blockade);                                
                                     double tempDistance1 = distance(edge2.getLine(), (Polygon)blockade1.getShape());
                                     double tempDistance2 = distance(edge3.getLine(), (Polygon)blockade1.getShape());
                                      if (tempDistance1<750) {
                                    	  blockededge1.add(blockade1);
                                      }
                                      if (tempDistance2<750) {
                                    	  blockededge2.add(blockade1);
                                      }
                                      if (tempDistance1<750 && tempDistance2<750) {
                                    	  return true; 
                                      }
                              }
                                  for (Blockade block1 : blockededge1) {
                                      for (Blockade block2 : blockededge2) {                           
                                          if (isblockPassable((Polygon)block1.getShape(), (Polygon)block2.getShape(), 750 ))
                                        	  return true;   
                                      }
                                  }
                          }
                      }
                  }
              }
          }
    	  }
    	return false;
    	
    }
    
    public  double distance(Polygon polygon, rescuecore2.misc.geometry.Point2D point) {
        if (polygon.contains(point.getX(), point.getY())) {
            return 0;
        }
        int count = polygon.npoints;
        double minDistance = Double.MAX_VALUE;
        for (int i = 0; i < count; i++) {
            int j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D stPoint = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[i], polygon.ypoints[i]);
            rescuecore2.misc.geometry.Point2D endPoint = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[j], polygon.ypoints[j]);
            rescuecore2.misc.geometry.Line2D poly2Line = new rescuecore2.misc.geometry.Line2D(stPoint, endPoint);
            double distance = distance(poly2Line, point);
            minDistance = Math.min(minDistance, distance);
            if (minDistance == 0.0) {
                break;
            }
        }
        return minDistance;
    }
   
 public  double distance(rescuecore2.misc.geometry.Line2D line, rescuecore2.misc.geometry.Point2D point) {
        return ptSegDist1(line.getOrigin().getX(), line.getOrigin().getY(), line.getEndPoint().getX(), line.getEndPoint().getY(),point.getX(), point.getY());
    }
   
 private double ptSegDist1 (double arg, double arg1, double arg3, double arg5, double arg7, double arg9) {
		arg3 -= arg;
		arg5 -= arg1;
		arg7 -= arg;
		arg9 -= arg1;
		double arg11 = arg7 * arg3 + arg9 * arg5;
		double arg13;
		if (arg11 <= 0.0D) {
			arg13 = 0.0D;
		} else {
			arg7 = arg3 - arg7;
			arg9 = arg5 - arg9;
			arg11 = arg7 * arg3 + arg9 * arg5;
			if (arg11 <= 0.0D) {
				arg13 = 0.0D;
			} else {
				arg13 = arg11 * arg11 / (arg3 * arg3 + arg5 * arg5);
			}
		}

		double arg15 = arg7 * arg7 + arg9 * arg9 - arg13;
		if (arg15 < 0.0D) {
			arg15 = 0.0D;
		}

		return Math.sqrt(arg15);
	}
   
 public boolean containsEach(Collection<Blockade> collection1, Collection<Blockade> collection2) {
        for (Object object : collection1) {
            if (collection2.contains(object)) {
                return true;
            }
        }
        return false;
    }
   
 public boolean isblockPassable(Polygon polygon, Polygon polygon1, int agentPassingThreshold) {

        int count = polygon1.npoints;
        int j;
        double tempDistance;
        boolean isPassable = false;
        for (int i = 0; i < count; i++) {
            j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D startPoint = new rescuecore2.misc.geometry.Point2D(polygon1.xpoints[i], polygon1.ypoints[i]);
            rescuecore2.misc.geometry.Point2D endPoint = new rescuecore2.misc.geometry.Point2D(polygon1.xpoints[j], polygon1.ypoints[j]);
            if (startPoint.equals(endPoint)) {
                continue;
            }
            rescuecore2.misc.geometry.Line2D poly2Line = new rescuecore2.misc.geometry.Line2D(startPoint, endPoint);
            tempDistance = distance(poly2Line, polygon);
            if (tempDistance < agentPassingThreshold) {
                isPassable = true;
                break;
            }
        }
        return isPassable;

    }
   
 public double distance(rescuecore2.misc.geometry.Line2D line, Polygon polygon) {
        List<rescuecore2.misc.geometry.Line2D> polyLines = getlines(polygon);
        double minDist = Double.MAX_VALUE;
        if (line.getEndPoint().equals(line.getOrigin())) {
            return distance(polygon, line.getOrigin());
        }
        for (rescuecore2.misc.geometry.Line2D polyLine : polyLines) {
            minDist = Math.min(minDist, distance(line, polyLine));
            if (minDist == 0.0) {
                break;
            }
        }
        return minDist;
    }

    public  List<rescuecore2.misc.geometry.Line2D> getlines(Polygon polygon) {
        List<rescuecore2.misc.geometry.Line2D> lines = new ArrayList<rescuecore2.misc.geometry.Line2D>();
        int count = polygon.npoints;
        for (int i = 0; i < count; i++) {
            int j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D p1 = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[i], polygon.ypoints[i]);
            rescuecore2.misc.geometry.Point2D p2 = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[j], polygon.ypoints[j]);
            rescuecore2.misc.geometry.Line2D line = new rescuecore2.misc.geometry.Line2D(p1, p2);
            lines.add(line);
        }
        return lines;
    }
  
    public  double distance(rescuecore2.misc.geometry.Line2D line1, rescuecore2.misc.geometry.Line2D line2) {
        if (linesIntersect(line1.getOrigin().getX(), line1.getOrigin().getY(), line1.getEndPoint().getX(), line1.getEndPoint().getY(),
                line2.getOrigin().getX(), line2.getOrigin().getY(), line2.getEndPoint().getX(), line2.getEndPoint().getY())) {
            return 0d;
        }
        double dist1 = distance(line1, line2.getOrigin());
        double dist2 = distance(line1, line2.getEndPoint());
        double dist3 = distance(line2, line1.getOrigin());
        double dist4 = distance(line2, line1.getEndPoint());
        double min = Math.min(dist1, dist2);
        min = Math.min(min, dist3);
        min = Math.min(min, dist4);
        return min;
    }
   
 public boolean linesIntersect(double arg, double arg1, double arg3, double arg5, double arg7, double arg9,
			double arg11, double arg13) {
		return relativeCCW(arg, arg1, arg3, arg5, arg7, arg9) * relativeCCW(arg, arg1, arg3, arg5, arg11, arg13) <= 0
				&& relativeCCW(arg7, arg9, arg11, arg13, arg, arg1)
						* relativeCCW(arg7, arg9, arg11, arg13, arg3, arg5) <= 0;
	}
    
    public int relativeCCW(double arg, double arg1, double arg3, double arg5, double arg7, double arg9) {
		arg3 -= arg;
		arg5 -= arg1;
		arg7 -= arg;
		arg9 -= arg1;
		double arg11 = arg7 * arg5 - arg9 * arg3;
		if (arg11 == 0.0D) {
			arg11 = arg7 * arg3 + arg9 * arg5;
			if (arg11 > 0.0D) {
				arg7 -= arg3;
				arg9 -= arg5;
				arg11 = arg7 * arg3 + arg9 * arg5;
				if (arg11 < 0.0D) {
					arg11 = 0.0D;
				}
			}
		}
		return arg11 < 0.0D ? -1 : (arg11 > 0.0D ? 1 : 0);
	}
   
    public List<EntityID> planMove(EntityID sourceArea, EntityID target) {

        List<EntityID> finalAreaPath = new ArrayList<>();

        if (target == null) {
            return finalAreaPath;
        }
        
        if (sourceArea.equals(target)) {
            System.out.println("Time:" + agentInfo.getTime() + " Already on target move =" + target);
            return finalAreaPath;
        }
        boolean repeatPlanning = repeatPlanning(target);
        boolean repeatAStar = !isPositionOnPreviousPath(sourceArea);
        if (repeatAStar || repeatPlanning) {           
            previousPath.clear();
            finalAreaPath = getShortestPath(sourceArea, target);
            previousTarget = target;
            previousPath = finalAreaPath;
        } else if (previousTarget.equals(target)) {
             ArrayList<EntityID> temp = new ArrayList<EntityID>();
            for (EntityID aPreviousPath : previousPath) {
                if (!sourceArea.equals(aPreviousPath)) {
                    temp.add(aPreviousPath);
                } else {
                    break;
                }
            }
            previousPath.removeAll(temp);
            finalAreaPath = previousPath;
        }
        return finalAreaPath;
    }
  
    private boolean repeatPlanning(EntityID target) {
        if (previousTarget == null || !previousTarget.equals(target) || continueMovingTime > 1) {
            continueMovingTime = 0;
            return true;
        } else {
            continueMovingTime++;
            return false;
        }
    }

    private boolean isPositionOnPreviousPath(EntityID position) {
        return previousPath.contains(position);
    }
}
