package SEU.module.complex;


import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.StandardMessage;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.*;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.BuildingDetector;
import SEU.module.algorithm.SEUFireClustering;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class SEUBuildingDetector extends BuildingDetector {
    private EntityID result;

    private int Mapsize;
    
    private Clustering clustering;
    private SEUFireClustering fireclustering; 
    									
    private Map<EntityID,EntityID> FireBrigadePosition=new HashMap<>();
    
    private int AverageArea;
    
    int testtimes=0;
    int testID=516746579;
    		
    public SEUBuildingDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        switch  (si.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.clustering = moduleManager.getModule("SEUBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                this.fireclustering=moduleManager.getModule("SEUBuildingDetector.SEUFireClustering", "SEU.module.algorithm.SEUFireClustering");
                break;
            case PRECOMPUTED:
                this.clustering = moduleManager.getModule("SEUBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                this.fireclustering=moduleManager.getModule("SEUBuildingDetector.SEUFireClustering", "SEU.module.algorithm.SEUFireClustering");
                break;
            case NON_PRECOMPUTE:
                this.clustering = moduleManager.getModule("SEUBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                this.fireclustering=moduleManager.getModule("SEUBuildingDetector.SEUFireClustering", "SEU.module.algorithm.SEUFireClustering");
                break;
        }
        for(StandardEntity entity:this.worldInfo.getEntitiesOfType(BUILDING))
        {
        	Building building=(Building) entity;
        	this.AverageArea+=building.getTotalArea()/10;  // to avoid overflow
        }
        Mapsize=(int) ((wi.getBounds().getHeight()+wi.getBounds().getWidth())/2);
        this.AverageArea=this.AverageArea*10/this.worldInfo.getEntitiesOfType(BUILDING).size();
        registerModule(this.clustering);
        registerModule(this.fireclustering);
    }

    @Override
    public BuildingDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.reflectMessage(messageManager);
        this.clustering.updateInfo(messageManager);
        this.fireclustering.updateInfo(messageManager);
        
        return this;
    }

    @Override
    public BuildingDetector calc() {
        //this.result = this.calcTargetInCluster();
        //if(this.result == null) {
            this.result = this.calcTargetInWorld();
        //}
        return this;
    }

    private EntityID calcTargetInCluster() {
        int clusterIndex = this.clustering.getClusterIndex(this.agentInfo.getID());
        Collection<StandardEntity> elements = this.clustering.getClusterEntities(clusterIndex);
        if(elements == null || elements.isEmpty()) {
            return null;
        }
        StandardEntity me = this.agentInfo.me();
        List<StandardEntity> fireBuildings = new ArrayList<>();
        List<StandardEntity> Edges=new ArrayList<>();
        for (StandardEntity entity : elements) {
            if (entity instanceof Building && ((Building)entity).isOnFire()) {
                fireBuildings.add(entity);
            }
        }
        for(StandardEntity entity:elements)
        {
        	if(entity instanceof Building)
        	{
        		int fireClusterIndex=this.fireclustering.getClusterIndex(entity);
        		if(fireClusterIndex>=0&& this.fireclustering.GetEdges(fireClusterIndex).contains(entity))
        			Edges.add(entity);
        	}
        }
        Edges.sort(new BuildingSorter(this.worldInfo,me,this.fireclustering,this.AverageArea));
        if(Edges.size()>0)
        	return Edges.get(0).getID();
        fireBuildings.sort(new BuildingSorter(this.worldInfo,me,this.fireclustering,this.AverageArea));
        if(fireBuildings.size()>0)
        	return fireBuildings.get(0).getID();
		return null;
		
    }

    private EntityID calcTargetInWorld() {
        Collection<StandardEntity> entities = this.worldInfo.getEntitiesOfType(
                StandardEntityURN.BUILDING,
                StandardEntityURN.GAS_STATION,
                StandardEntityURN.AMBULANCE_CENTRE,
                StandardEntityURN.FIRE_STATION,
                StandardEntityURN.POLICE_OFFICE
        );
        StandardEntity me = this.agentInfo.me();
        List<StandardEntity> fireBuildings = new ArrayList<>();
        List<StandardEntity> Edges=new ArrayList<>();
        
        for(List<StandardEntity> list:this.fireclustering.GetAllEdges())
        {
        	for(StandardEntity entity:list)
        	{
        		Edges.add(entity);
        	}
        }
        //System.out.println("Edges size="+Edges.size()+" MYid="+this.agentInfo.getID().getValue());
        for (StandardEntity entity : entities) {
            if (((Building)entity).isOnFire()) {
                fireBuildings.add(entity);
            }
        }
        
     
        if(Edges.size()>0)
        {
        	Edges.sort(new BuildingSorter(this.worldInfo,me,this.fireclustering,this.AverageArea));
        	return Edges.get(0).getID();
        }
        //fireBuildings.sort(new BuildingSorter(this.worldInfo,me,this.fireclustering,this.FireBrigadeTask,this.AverageArea));
        fireBuildings.sort(new DistanceSorter(this.worldInfo,this.agentInfo.me()));
        if(fireBuildings.size()>0)
        	return fireBuildings.get(0).getID();
		return null;
			
    }

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public BuildingDetector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.clustering.precompute(precomputeData);
        return this;
    }

    @Override
    public BuildingDetector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.clustering.resume(precomputeData);
        return this;
    }

    @Override
    public BuildingDetector preparate() {
        super.preparate();
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.clustering.preparate();
        return this;
    }

    

    

    private class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;

        DistanceSorter(WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
        }

        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = this.worldInfo.getDistance(this.reference, a);
            int d2 = this.worldInfo.getDistance(this.reference, b);
            return d1 - d2;
        }
    }
    
    private class BuildingSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;
        private SEUFireClustering firecluster;
        private List<StandardEntity> fireagents=new ArrayList<>();
        private int AverageArea;
        BuildingSorter(WorldInfo wi, StandardEntity reference,SEUFireClustering firecluster,int Average) {
            this.reference = reference;
            this.worldInfo = wi;
            this.AverageArea=Average;
            this.firecluster=firecluster;
            for(StandardEntity entity:wi.getEntitiesOfType(FIRE_BRIGADE))
            {
            	fireagents.add(entity);
            }
        }

        public int compare(StandardEntity a, StandardEntity b) {
        	FireBrigade me=(FireBrigade) reference;
            int Acost=0,Bcost=0;
            int DistanceCostofA=0,NotFireBuildingsAround=0,ToCenterCostofA=0,AroundCostofA=0,GasCostofA=500000,TempCostofA=0;
            int DistanceCostofB=0,NotFireBuildingsBround=0,ToCenterCostofB=0,AroundCostofB=0,GasCostofB=500000,TempCostofB=0;
            int RankCostofA=0;
            int RankCostofB=0;
            int Ax=0,Ay=0,AclusterIndex=this.firecluster.getClusterIndex(a.getID());
            int Bx=0,By=0,BclusterIndex=this.firecluster.getClusterIndex(b.getID());
         
            int Distancethreshold=25000,price=280,Rankprice=Mapsize/20,GasPrice=Rankprice*2,Tempcoe=40,GasDisthreshold=15000;
            
            Building buildingA=(Building) a;
            Building buildingB=(Building) b;
            
            TempCostofA=buildingA.getTemperature()*buildingA.getFieryness()*Tempcoe;
            TempCostofB=buildingB.getTemperature()*buildingB.getFieryness()*Tempcoe;
            
            double AneedAgents=buildingA.getTotalArea()/this.AverageArea+1;
            AneedAgents=Math.min(AneedAgents, this.fireagents.size());
            double BneedAgents=buildingB.getTotalArea()/this.AverageArea+1;
            BneedAgents=Math.min(BneedAgents, this.fireagents.size());
            
            Ax=buildingA.getX();Ay=buildingA.getY();
            Bx=buildingB.getX();By=buildingB.getY();
            
            if(AclusterIndex!=-1&&BclusterIndex!=-1)
            {
            	  Pair<Integer,Integer> Acenter=this.firecluster.GetCenter(AclusterIndex);
                  Pair<Integer,Integer> Bcenter=this.firecluster.GetCenter(BclusterIndex);
                  ToCenterCostofA=(int) Math.hypot(Acenter.first()-Ax, Acenter.second()-Ay);
                  ToCenterCostofB=(int) Math.hypot(Bcenter.first()-Bx, Bcenter.second()-By);
            }
            
            fireagents.sort(new DistanceSorter(this.worldInfo,a));
            RankCostofA=fireagents.indexOf(me)*Rankprice;
            fireagents.sort(new DistanceSorter(this.worldInfo,b));
            RankCostofB=fireagents.indexOf(me)*Rankprice;
            
            DistanceCostofA=this.worldInfo.getDistance(me.getID(), a.getID());
            DistanceCostofB=this.worldInfo.getDistance(me.getID(), b.getID());
            
            for(EntityID entityid:this.worldInfo.getObjectIDsInRange(a.getID(), Distancethreshold))
            {
            	if(this.worldInfo.getEntity(entityid) instanceof Building)
            	{
            		Building building=(Building) this.worldInfo.getEntity(entityid);
            		if(!building.isFierynessDefined())
            			NotFireBuildingsAround+=2;
            		else
            		{
            			if(building.getFieryness()==0)
                			NotFireBuildingsAround+=4;
                		else if(building.getFieryness()<=7&&building.getFieryness()>=4)
                			NotFireBuildingsAround+=1;
            		}
            	}
            }
            
            for(EntityID entityid:this.worldInfo.getObjectIDsInRange(b.getID(), Distancethreshold))
            {
            	if(this.worldInfo.getEntity(entityid) instanceof Building)
            	{
            		Building building=(Building) this.worldInfo.getEntity(entityid);
            		if(!building.isFierynessDefined())
            			NotFireBuildingsBround+=2;
            		else
            		{
            			if(building.getFieryness()==0)
                			NotFireBuildingsBround+=4;
                		else if(building.getFieryness()<=7&&building.getFieryness()>=4)
                			NotFireBuildingsBround+=1;
            		}
            	}
            } 
            List<StandardEntity> Gasstation=new ArrayList<>();
            for(StandardEntity entity:this.worldInfo.getEntitiesOfType(GAS_STATION))
            {
            	Gasstation.add(entity);
            }
            Gasstation.sort(new DistanceSorter(this.worldInfo,a));
            if(Gasstation.size()>0)
            {
            	if(this.worldInfo.getObjectIDsInRange(Gasstation.get(0).getID(), GasDisthreshold).contains(a.getID()))
            	{GasCostofA=this.worldInfo.getDistance(Gasstation.get(0).getID(), a.getID());}
                Gasstation.sort(new DistanceSorter(this.worldInfo,b));
                if(this.worldInfo.getObjectIDsInRange(Gasstation.get(0).getID(), GasDisthreshold).contains(b.getID()))
                {GasCostofB=this.worldInfo.getDistance(Gasstation.get(0).getID(), b.getID());}
            }
            
            
            AroundCostofA=NotFireBuildingsAround*price;
            AroundCostofB=NotFireBuildingsBround*price;
        
            Acost=DistanceCostofA+TempCostofA+5*GasCostofA+RankCostofA-AroundCostofA-ToCenterCostofA;
            Bcost=DistanceCostofB+TempCostofB+5*GasCostofB+RankCostofB-AroundCostofB-ToCenterCostofB;
            if(a.equals(b))
        		return 0;
            return Acost - Bcost;
        }
    }
    private void reflectMessage(MessageManager messageManager) {
        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
        changedEntities.add(this.agentInfo.getID());
        int time = this.agentInfo.getTime();

        for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
            Class<? extends CommunicationMessage> messageClass = message.getClass();
            if(messageClass == MessageBuilding.class) {
                MessageBuilding mb = (MessageBuilding)message;
                if(!changedEntities.contains(mb.getBuildingID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mb);
                }
            } else if(messageClass == MessageRoad.class) {
                MessageRoad mr = (MessageRoad)message;
                if(mr.isBlockadeDefined() && !changedEntities.contains(mr.getBlockadeID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mr);
                }
                
            } else if(messageClass == MessageCivilian.class) {
                MessageCivilian mc = (MessageCivilian) message;
                if(!changedEntities.contains(mc.getAgentID())){
                    MessageUtil.reflectMessage(this.worldInfo, mc);
                }
                
            } else if(messageClass == MessageAmbulanceTeam.class) {
                MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;
                if(!changedEntities.contains(mat.getAgentID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mat);
                }
               
            } else if(messageClass == MessageFireBrigade.class) {
                MessageFireBrigade mfb = (MessageFireBrigade) message;
                if(!changedEntities.contains(mfb.getAgentID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mfb);
                }
               
            } else if(messageClass == MessagePoliceForce.class) {
                MessagePoliceForce mpf = (MessagePoliceForce) message;
                if(!changedEntities.contains(mpf.getAgentID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mpf);
                }
                
            }
        }
    }
}
