package SEU.extaction;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.action.police.ActionClear;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;
import com.google.common.collect.Lists;
import rescuecore2.config.NoSuchConfigOptionException;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;
import java.util.stream.Collectors;

public class ActionExtClear extends ExtAction {
    private PathPlanning pathPlanning;

    private int clearRadius;
    private int clearDistance;
    private int forcedMove;
    private int thresholdRest;
    private int kernelTime;

    private EntityID target;
    private int oldClearX;
    private int oldClearY;
    private int count;
    private int lastLocaX;
    private int lastLocaY;
    private int lastDestX;
    private int lastDestY;
    private boolean needContinue;
    private int lastTime;

    public ActionExtClear(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        this.clearRadius = si.getClearRepairRad();
        this.clearDistance = si.getClearRepairDistance();
        this.forcedMove = developData.getInteger("ActionExtClear.forcedMove", 3);
        this.thresholdRest = developData.getInteger("ActionExtClear.rest", 100);

        this.target = null;
        this.oldClearX = 0;
        this.oldClearY = 0;
        this.count = 0;
        this.lastDestX = 0;
        this.lastDestY = 0;
        this.lastLocaX = 0;
        this.lastLocaY = 0;
        this.needContinue = false;
        this.lastTime = -1;

        switch  (si.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
        }
    }

    @Override
    public ExtAction precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.pathPlanning.precompute(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.pathPlanning.resume(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction updateInfo(MessageManager messageManager){
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);

        if (this.agentInfo.getTime() - this.lastTime > 3) {
            this.needContinue = false;
        }
        this.lastTime = this.agentInfo.getTime();

        return this;
    }

    @Override
    public ExtAction setTarget(EntityID target) {
        this.target = null;
        StandardEntity entity = this.worldInfo.getEntity(target);
        if(entity != null) {
            if (entity instanceof Road) {
                this.target = target;
            } else if (entity.getStandardURN().equals(StandardEntityURN.BLOCKADE)) {
                this.target = ((Blockade) entity).getPosition();
            } else if (entity instanceof Building) {
                this.target = target;
            }
        }
        return this;
    }

    @Override
    public ExtAction calc() {
        this.result = null;
        PoliceForce policeForce = (PoliceForce) this.agentInfo.me();
        if (this.needRest(policeForce)) {
            List<EntityID> list = new ArrayList<>();
            if (this.target != null) {
                list.add(this.target);
            }
            this.result = this.calcRest(policeForce, this.pathPlanning, list);
            if (this.result != null) {
                this.lastLocaX = policeForce.getX();
                this.lastLocaY = policeForce.getY();
                return this;
            }
        }

        if (this.target == null) {
            return this;
        }
        EntityID agentPosition = policeForce.getPosition();
        StandardEntity targetEntity = this.worldInfo.getEntity(this.target);
        StandardEntity positionEntity = Objects.requireNonNull(this.worldInfo.getEntity(agentPosition));
        if (targetEntity == null || !(targetEntity instanceof Area)) {
            return this;
        }
        if (this.needContinue) { //if last target is not completed, continue to do it
            this.result = this.getContinueAction(policeForce);
            if (this.result != null) {
                this.lastLocaX = policeForce.getX();
                this.lastLocaY = policeForce.getY();
                return this;
            }
        }
        if (agentPosition.equals(this.target)) {
            this.result = this.getAreaClearAction(policeForce);
        } else if (((Area) targetEntity).getEdgeTo(agentPosition) != null) {
            this.result = this.getNeighbourPositionAction(policeForce, (Area) targetEntity);
        } else {
            List<EntityID> path = this.pathPlanning.getResult(agentPosition, this.target);
            if (path != null && path.size() > 0) {
                int index = path.indexOf(agentPosition);
                if (index == -1) {
                    Area area = (Area) positionEntity;
                    for (int i = 0; i < path.size(); i++) {
                        if (area.getEdgeTo(path.get(i)) != null) {
                            index = i;
                            break;
                        }
                    }
                } else if (index >= 0) {
                    index++;
                }
                if (index >= 0 && index < (path.size())) {
                    StandardEntity entity = this.worldInfo.getEntity(path.get(index));
                    this.result = this.getNeighbourPositionAction(policeForce, (Area) entity);
                    if (this.result != null && this.result.getClass() == ActionMove.class) {
                        if (!((ActionMove) this.result).getUsePosition()) {
                            this.result = null;
                        }
                    }
                }
                if (this.result == null) {
                    this.result = new ActionMove(path);
                    this.count = 0;
                }
            }
        }
        this.lastLocaX = policeForce.getX();
        this.lastLocaY = policeForce.getY();
        return this;
    }

    @Override
    public Action getAction() {
        return this.result;
    }

    private Action getContinueAction(PoliceForce police) { //basic function
        if (!needContinue) {
            this.lastDestX = 0;
            this.lastDestY = 0;
            return null;
        }
        this.needContinue = false;
        EntityID position = police.getPosition();
        StandardEntity entity = this.worldInfo.getEntity(position);
        if (!(entity instanceof Road)) {
            this.lastDestX = 0;
            this.lastDestY = 0;
            return null;
        }
        Road road = (Road)entity;
        if (!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
            this.lastDestX = 0;
            this.lastDestY = 0;
            return null;
        }
        this.needContinue = true;
        Collection<Blockade> blockades = this.worldInfo.getBlockades(road);
        int policeX = police.getX();
        int policeY = police.getY();
        Set<Blockade> covers = new HashSet<>();
        for (Blockade blockade : blockades) {
            if (this.isInside(policeX, policeY, blockade)) {
                covers.add(blockade);
            }
        }
        if (!covers.isEmpty()) { //if police is covered by blockade, clear it first
            return new ActionClear((Blockade) this.getClosestEntity(covers, this.agentInfo.me()));
        }
        covers.clear();
        Vector2D vector = (new Vector2D(this.lastDestX - policeX, this.lastDestY - policeY))
                .normalised().scale(this.clearDistance);
        int clearX = policeX + (int) vector.getX();
        int clearY = policeY + (int) vector.getY();
        vector = vector.normalised().scale(-300);
        int startX = policeX + (int) vector.getX();
        int startY = policeY + (int) vector.getY();
        vector = vector.normalised().scale(250).getNormal();
        List<Blockade> removeBlockadesList = new ArrayList<>();
        int farX = this.getDistance(policeX, policeY, this.lastDestX, this.lastDestY) >
                this.getDistance(policeX, policeY, clearX, clearY) ? this.lastDestX : clearX;
        int farY = farX == clearX ? clearY : this.lastDestY;
        for (Blockade blockade : blockades) { //functions in java.awt.geom.Area are not fast enough
            if (!(this.intersect(startX, startY, farX, farY, blockade) ||
                    this.intersect(policeX, policeY, farX + vector.getX(), farY + vector.getY(), blockade) ||
                    this.intersect(policeX, policeY, farX - vector.getX(), farY - vector.getY(), blockade) ||
                    this.intersect(policeX, policeY, startX + vector.getX(), startY + vector.getY(), blockade) ||
                    this.intersect(policeX, policeY, startX - vector.getX(), startY - vector.getY(), blockade))) {
                removeBlockadesList.add(blockade);
            }
        }
        blockades.removeAll(removeBlockadesList);
        removeBlockadesList.clear();
        if (blockades.isEmpty()) {
            this.lastDestX = 0;
            this.lastDestY = 0;
            this.needContinue = false;
            return null;
        }
        Blockade closest = (Blockade)this.getClosestEntity(blockades, police);
        if (!(this.intersect(startX, startY, clearX, clearY, closest) ||
                this.intersect(startX - vector.getX(), startY - vector.getY(),
                        clearX + vector.getX(), clearY + vector.getY(), closest) ||
                this.intersect(startX + vector.getX(), startY + vector.getY(),
                        clearX - vector.getX(), clearY - vector.getY(), closest))) {
            this.oldClearX = 0;
            this.oldClearY = 0;
            this.count = 0;
            return new ActionMove(Lists.newArrayList(position), lastDestX, lastDestY);
        } else {
            if (this.equalsPoint(clearX, clearY, this.oldClearX, this.oldClearY, 1000)) {
                ++this.count;
                if (this.count >= this.forcedMove) {
                    this.oldClearX = 0;
                    this.oldClearY = 0;
                    this.count = 0;
                    return new ActionMove(Lists.newArrayList(position), clearX, clearY);
                }
            }
            this.oldClearX = clearX;
            this.oldClearY = clearY;
            return new ActionClear(clearX, clearY);
        }
    }

    private Action getRescueAction(PoliceForce police, Edge nextEgde) {
        EntityID position = police.getPosition();
        StandardEntity entity = this.worldInfo.getEntity(position);
        if (!(entity instanceof Road)) {
            return null;
        }
        Road road = (Road)entity;
        if (!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
            return null;
        }
        Collection<Blockade> blockades = this.worldInfo.getBlockades(road);
        Set<Human> onRoad = new HashSet<>(); //get all other kinds of human in this area
        for (StandardEntity e : this.worldInfo.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM,
                StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.CIVILIAN)) {
            Human h = (Human)e;
            if (h.isPositionDefined() && h.getPosition().equals(position) && h.isXDefined() && h.isYDefined()) {
                onRoad.add(h);
            }
        }
        Set<Human> stucked = new HashSet<>();
        for (Human human : onRoad) { //find those who were stucked by blockades
            int hX = human.getX();
            int hY = human.getY();
            for (Blockade blockade : blockades) {
                if (this.isInside(hX, hY, blockade) || this.isNearBlockade(hX, hY, blockade, 500)) {
                    stucked.add(human);
                    break;
                }
            }
        }
        onRoad.clear();
        if (stucked.isEmpty()) {
            return null;
        }
        int policeX = police.getX();
        int policeY = police.getY();
        Set<Human> byTheWay = new HashSet<>(); //find those who can be helped when clearing toward the next edge
        if (nextEgde != null) {
            Point2D mid = this.getMidPoint(nextEgde.getLine());
            int midX = (int)mid.getX();
            int midY = (int)mid.getY();
            for (Human human : stucked) {
                if (java.awt.geom.Line2D.ptLineDist(policeX, policeY, midX, midY, human.getX(), human.getY()) <
                        this.clearRadius*4/5) {
                    byTheWay.add(human);
                }
            }
            stucked.removeAll(byTheWay);
        }
        if (!stucked.isEmpty()) { //help the rest of human first
            byTheWay.clear();
            int bestTargetX = 0;
            int bestTargetY = 0;
            int count = -1;
            for (Human human : stucked) { //find a direction that can help most human by one action
                int humanX = human.getX();
                int humanY = human.getY();
                Vector2D vector = new Vector2D(humanX - policeX, humanY - policeY);
                if (vector.getLength() < this.clearDistance) {
                    vector.normalised().scale(this.clearDistance + 500);
                }
                humanX = policeX + (int)vector.getX();
                humanY = policeY + (int)vector.getY();
                int counter = 0;
                for (Human other : stucked) {
                    if (other.equals(human)) {
                        continue;
                    }
                    if (java.awt.geom.Line2D.ptLineDist(humanX, humanY, policeX, policeY, other.getX(), other.getY())
                            < this.clearRadius * 4 / 5) {
                        ++counter;
                    }
                }
                if (counter > count) {
                    count = counter;
                    bestTargetX = humanX;
                    bestTargetY = humanY;
                } else if (counter == count) {
                    if (this.worldInfo.getDistance(police,human) >
                            this.getDistance(policeX, policeY,bestTargetX, bestTargetY)) {
                        bestTargetX = humanX;
                        bestTargetY = humanY;
                    }
                }
            }
            this.lastDestX = bestTargetX;
            this.lastDestY = bestTargetY;
            this.needContinue = true;
            return this.getContinueAction(police);
        }
        if (!byTheWay.isEmpty()) { //then clear and move to the next edge
            Point2D point = this.getMidPoint(nextEgde.getLine());
            this.lastDestX = (int)point.getX();
            this.lastDestY = (int)point.getY();
            this.needContinue = true;
            return this.getContinueAction(police);
        }
        return null; //impossible branch
    }

    private Action getAreaClearAction(PoliceForce police) {
        EntityID position = police.getPosition();
        StandardEntity entity = this.worldInfo.getEntity(position);
        if (!(entity instanceof Road)) {
            return null;
        }
        Road road = (Road)entity;
        if (!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
            return null;
        }
        Action action = this.getRescueAction(police, null); //help human first
        if (action != null) {
            return action;
        }
        List<Edge> passableEdge = new ArrayList<>(this.getPassableEdge(road));
        passableEdge.sort(new DistanceEdgeSorter(new Point2D(police.getX(), police.getY())));
        for (int i = passableEdge.size() - 1; i >= 0; --i) { //then choose farthest edge that need clear
            Point2D point = this.getMidPoint(passableEdge.get(i).getLine());
            this.lastDestX = (int)point.getX();
            this.lastDestY = (int)point.getY();
            this.needContinue = true;
            action = this.getContinueAction(police);
            if (action != null) {
                return action;
            }
        }
        int rX = road.getX(); //make sure that the central of the road is wide enough for human to walk through
        int rY = road.getY();
        Collection<Blockade> blockades = this.worldInfo.getBlockades(road);
        for (Edge edge : passableEdge) {
            Point2D mid = this.getMidPoint(edge.getLine());
            double mX = mid.getX();
            double mY = mid.getY();
            Vector2D vector = (new Vector2D(rX - mX, rY - mY)).normalised().scale(250).getNormal();
            for (Blockade blockade : blockades) {
                if (this.intersect(rX, rY, mX, mY, blockade) ||
                        this.intersect(rX + vector.getX(), rY + vector.getY(),
                                mX + vector.getX(), mY + vector.getY(), blockade) ||
                        this.intersect(rX - vector.getX(), rY - vector.getY(),
                                mX - vector.getX(), mY - vector.getY(), blockade)) {
                    return new ActionClear(blockade);
                }
            }
        }
        return null;
    }

    private Action getNeighbourPositionAction(PoliceForce police, Area target) {
        double agentX = police.getX();
        double agentY = police.getY();
        StandardEntity position = Objects.requireNonNull(this.worldInfo.getPosition(police));
        Edge edge = ((Area)position).getEdgeTo(target.getID());
        if(edge == null) {
            return null;
        }
        if(position instanceof Road) {
            Action action = this.getRescueAction(police, edge);
            if (action != null) { //help human first
                return  action;
            } //if there is no human to help, clear and move to the edge
            Point2D point = this.getMidPoint(edge.getLine());
            this.lastDestX = (int)point.getX();
            this.lastDestY = (int)point.getY();
            this.needContinue = true;
            action = this.getContinueAction(police);
            if (action != null) {
                return action;
            }
        }
        if(target instanceof Road) {
            Road road = (Road)target;
            if(!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
                return new ActionMove(Lists.newArrayList(position.getID(), target.getID()));
            }

            double rX = road.getX();
            double rY = road.getY();
            Point2D point = this.getMidPoint(edge.getLine());
            double mX = point.getX();
            double mY = point.getY();
            if (this.getDistance(rX, rY, mX, mY) > this.clearDistance) {
                Vector2D vector = (new Vector2D(rX - mX, rY - mY)).normalised().scale(this.clearDistance/2);
                rX = mX + vector.getX();
                rY = mY + vector.getY();
            }
            Vector2D vector = (new Vector2D(mX - rX, mY - rY)).normalised().scale(250).getNormal();
            Collection<Blockade> blockades = this.worldInfo.getBlockades(road);
            boolean flag = false;
            for (Blockade blockade : blockades) {
                if (this.intersect(mX, mY, rX, rY, blockade) ||
                        this.intersect(mX + vector.getX(), mY + vector.getY(),
                                rX + vector.getX(), rY + vector.getY(), blockade) ||
                        this.intersect(mX - vector.getX(), mY - vector.getY(),
                                rX - vector.getX(), rY - vector.getY(), blockade)) {
                    flag = true;
                    break;
                }
            }
            if (!flag) { //if there is enough place for police in the next road, move to it
                this.count = 0;
                this.oldClearX = 0;
                this.oldClearY = 0;
                return new ActionMove(Lists.newArrayList(position.getID(), target.getID()));
            }

            Blockade clearBlockade = null;
            Double minPointDistance = Double.MAX_VALUE;
            int clearX = 0;
            int clearY = 0;
            for (Blockade blockade : blockades) {
                if (blockade != null && blockade.isApexesDefined()) {
                    int[] apexes = blockade.getApexes();
                    for (int i = 0; i < (apexes.length - 2); i += 2) {
                        double distance = this.getDistance(agentX, agentY, apexes[i], apexes[i + 1]);
                        if (distance < minPointDistance) {
                            clearBlockade = blockade;
                            minPointDistance = distance;
                            clearX = apexes[i];
                            clearY = apexes[i + 1];
                        }
                    }
                }
            }
            if (clearBlockade != null && minPointDistance < this.clearDistance) {
                vector = (new Vector2D(clearX - agentX, clearY - agentY)).normalised().scale(this.clearDistance);
                clearX = (int) (agentX + vector.getX());
                clearY = (int) (agentY + vector.getY());
                if(this.equalsPoint(this.oldClearX, this.oldClearY, clearX, clearY, 1000)) {
                    if(this.count >= this.forcedMove) {
                        this.count = 0;
                        this.oldClearX = 0;
                        this.oldClearY = 0;
                        return new ActionMove(Lists.newArrayList(position.getID(), road.getID()));
                    }
                    this.count++;
                } else if (this.equalsPoint(this.lastLocaX, this.lastLocaY, agentX, agentY, 1000)) {
                    if (this.count >= this.forcedMove) {
                        this.count = 0;
                        this.oldClearX = 0;
                        this.oldClearY = 0;
                        return new ActionMove(Lists.newArrayList(position.getID(), road.getID()));
                    }
                    this.count++;
                }
                this.oldClearX = clearX;
                this.oldClearY = clearY;
                return new ActionClear(clearX, clearY);
            }
        }
        this.count = 0;
        this.oldClearX = 0;
        this.oldClearY = 0;
        return new ActionMove(Lists.newArrayList(position.getID(), target.getID()));
    }

    private Action calcRest(Human human, PathPlanning pathPlanning, Collection<EntityID> targets) {
        EntityID position = human.getPosition();
        Collection<EntityID> refuges = this.worldInfo.getEntityIDsOfType(StandardEntityURN.REFUGE);
        int currentSize = refuges.size();
        if(refuges.contains(position)) {
            return new ActionRest();
        }
        List<EntityID> firstResult = null;
        while(refuges.size() > 0) {
            pathPlanning.setFrom(position);
            pathPlanning.setDestination(refuges);
            List<EntityID> path = pathPlanning.calc().getResult();
            if (path != null && path.size() > 0) {
                if (firstResult == null) {
                    firstResult = new ArrayList<>(path);
                    if(targets == null || targets.isEmpty()) {
                        break;
                    }
                }
                EntityID refugeID = path.get(path.size() - 1);
                pathPlanning.setFrom(refugeID);
                pathPlanning.setDestination(targets);
                List<EntityID> fromRefugeToTarget = pathPlanning.calc().getResult();
                if (fromRefugeToTarget != null && fromRefugeToTarget.size() > 0) {
                    return new ActionMove(path);
                }
                refuges.remove(refugeID);
                //remove failed
                if (currentSize == refuges.size()) {
                    break;
                }
                currentSize = refuges.size();
            } else {
                break;
            }
        }
        return firstResult != null ? new ActionMove(firstResult) : null;
    }

    private boolean equalsPoint(double p1X, double p1Y, double p2X, double p2Y, double range) {
        return (p2X - range < p1X && p1X < p2X + range) && (p2Y - range < p1Y && p1Y < p2Y + range);
    }

    private boolean isInside(double pX, double pY, Blockade blockade) {
        Point2D p = new Point2D(pX, pY);
        int apex[] = blockade.getApexes();
        Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
        Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
        double theta = this.getAngle(v1, v2);

        for(int i = 0; i < apex.length - 2; i += 2) {
            v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
            v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
            theta += this.getAngle(v1, v2);
        }
        return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
    }

    private double getAngle(Vector2D v1, Vector2D v2) {
        double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
        double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
        if(flag > 0) {
            return angle;
        }
        if(flag < 0) {
            return -1 * angle;
        }
        return 0.0D;
    }

    private boolean isNearBlockade(double pX, double pY, Blockade blockade, double range) {
        int[] apex = blockade.getApexes();
        for (int i = 0; i < apex.length - 4; i += 2) {
            if(java.awt.geom.Line2D.ptLineDist(apex[i], apex[i + 1], apex[i + 2], apex[i + 3], pX, pY) < range) {
                return true;
            }
        }
        if (java.awt.geom.Line2D.ptLineDist(apex[0], apex[1], apex[apex.length - 2], apex[apex.length - 1], pX, pY) < range) {
            return true;
        }
        return false;
    }

    private boolean intersect(double agentX, double agentY, double pointX, double pointY, Blockade blockade) {
        List<Line2D> lines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(blockade.getApexes()), true);
        for(Line2D line : lines) {
            Point2D start = line.getOrigin();
            Point2D end = line.getEndPoint();
            double startX = start.getX();
            double startY = start.getY();
            double endX = end.getX();
            double endY = end.getY();
            if(java.awt.geom.Line2D.linesIntersect(
                    agentX, agentY, pointX, pointY,
                    startX, startY, endX, endY
            )) {
                return true;
            }
        }
        return false;
    }

    private double getDistance(double fromX, double fromY, double toX, double toY) {
        return Math.hypot(toX - fromX, toY - fromY);
    }

    private Point2D getMidPoint(Line2D line) {
        return new Point2D((line.getOrigin().getX() + line.getEndPoint().getX()) / 2,
                (line.getOrigin().getY() + line.getEndPoint().getY()) / 2);
    }

    private Set<Edge> getPassableEdge(Area area) {
        Set<Edge> passableEdges = new HashSet<>();
        for (Edge edge : area.getEdges()) {
            if (edge.isPassable()) {
                passableEdges.add(edge);
            }
        }
        return passableEdges;
    }

    private boolean needRest(Human agent) {
        int hp = agent.getHP();
        int damage = agent.getDamage();
        if(damage == 0 || hp == 0) {
            return false;
        }
        int activeTime = (hp / damage) + ((hp % damage) != 0 ? 1 : 0);
        if(this.kernelTime == -1) {
            try {
                this.kernelTime = this.scenarioInfo.getKernelTimesteps();
            }catch (NoSuchConfigOptionException e) {
                this.kernelTime = -1;
            }
        }
        return damage >= this.thresholdRest || (activeTime + this.agentInfo.getTime()) < this.kernelTime;
    }

    private StandardEntity getClosestEntity(Collection<? extends StandardEntity> entities, StandardEntity reference) {
        if (entities.isEmpty()) {
            return null;
        }
        double minDistance = Double.MAX_VALUE;
        StandardEntity closestEntity = null;
        for (StandardEntity entity : entities) {
            double distance = this.worldInfo.getDistance(reference, entity);
            if (distance < minDistance) {
                minDistance = distance;
                closestEntity = entity;
            }
        }
        return closestEntity;
    }

    private class DistanceEdgeSorter implements Comparator<Edge> {
        private Point2D reference;

        DistanceEdgeSorter(Point2D reference) {
            this.reference = reference;
        }

        public int compare(Edge a, Edge b) {
            int d1 = (int) java.awt.geom.Line2D.ptLineDist(a.getStartX(), a.getStartY(), a.getEndX(), a.getEndY(), reference.getX(), reference.getY());
            int d2 = (int) java.awt.geom.Line2D.ptLineDist(b.getStartX(), b.getStartY(), b.getEndX(), b.getEndY(), reference.getX(), reference.getY());
            return d1 - d2;
        }
    }
}
