package SEU.extaction;


import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.action.fire.ActionExtinguish;
import adf.agent.action.fire.ActionRefill;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;
import SEU.module.algorithm.SEUFireClustering;
import rescuecore2.config.NoSuchConfigOptionException;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.HYDRANT;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;

public class ActionFireFighting extends ExtAction {
    private PathPlanning pathPlanning;
   
    private int maxExtinguishDistance;
    private int maxExtinguishPower;
    private int thresholdRest;
    private int kernelTime;
    private int refillCompleted;
    private int refillRequest;
    
    private boolean refillFlag;
    private boolean StuckFlag;
    
    private EntityID target;
   
    //------------------------------------------------------------
    private SEUFireClustering fireclustering;
    private int ExtinguishPower; 	
    private EntityID refilltarget;
 
    private EntityID lastposition=null;
    
    //CrossRoad informations
    private List<StandardEntity> cross=new ArrayList<>();
    private List<StandardEntity> removedcross=new ArrayList<>();
    private List<StandardEntity> cangocross=new ArrayList<>();
    
    //Judge whether an agent was surrounded by fires
    private boolean shouldIgo=false;
    //Means I escaped from fires
    private boolean goingout=false;
    //Means when I escaped,I have no target to extinguish
    private boolean goingback=false;
    
    private int myclusterIndex=-1;
    
    //Entrance roads
    private List<StandardEntity> Entrances=new ArrayList<>();
    
    //Informations about the water quantity of all the FireBrigade( in my worldInfo )
    private Map<EntityID,Integer> firebrigadeWater=new HashMap<>();
    //Store all the roads that can't see a certain Building
    private Map<EntityID,List<StandardEntity>> WrongPosition=new HashMap<>();
    //A flag to show whether a map contains fire station in order to use different strategy
    private int Mode=0;
    
    //for TEST
    private int testID=588570422;
    
    //Extinguish water coefficient
    private int AreaCoe=10;  
    private int TempCoe=15;
    
    private Pair<Integer,Integer> Walkdire;
    
    //-------------------------------------------------------------
    public ActionFireFighting(AgentInfo agentInfo, WorldInfo worldInfo, ScenarioInfo scenarioInfo, ModuleManager moduleManager, DevelopData developData) {
        super(agentInfo, worldInfo, scenarioInfo, moduleManager, developData);
        this.maxExtinguishDistance = scenarioInfo.getFireExtinguishMaxDistance();
        this.maxExtinguishPower = scenarioInfo.getFireExtinguishMaxSum();
        this.thresholdRest = developData.getInteger("ActionFireFighting.rest", 100);
        int maxWater = scenarioInfo.getFireTankMaximum();
        this.refillCompleted = (maxWater / 10) * developData.getInteger("ActionFireFighting.refill.completed", 10);
        this.refillRequest = this.maxExtinguishPower * developData.getInteger("ActionFireFighting.refill.request", 1);
        this.refillFlag = false;
        
        this.target = null;
        this.refilltarget=null;
        
        switch  (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("ActionFireFighting.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.fireclustering=moduleManager.getModule("ActionFireFighting.SEUFireClustering", "SEU.module.algorithm.SEUFireClustering");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("ActionFireFighting.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.fireclustering=moduleManager.getModule("ActionFireFighting.SEUFireClustering", "SEU.module.algorithm.SEUFireClustering");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("ActionFireFighting.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.fireclustering=moduleManager.getModule("ActionFireFighting.SEUFireClustering", "SEU.module.algorithm.SEUFireClustering");
                break;
        }
        this.FindCross();
        this.cangocross=cross;
        if(worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_STATION).size()!=0)
        	this.Mode=1;
        
        for(StandardEntity entity:this.worldInfo.getEntitiesOfType(StandardEntityURN.BUILDING))
        {
        	Building building=(Building) entity;
        	for(EntityID entityid:building.getNeighbours())
        	{
        		if(this.worldInfo.getEntity(entityid) instanceof Road)
        		{
        			this.Entrances.add(this.worldInfo.getEntity(entityid));
        		}
        	}
        }
    }

    @Override
    public ExtAction precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.pathPlanning.precompute(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.pathPlanning.resume(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction updateInfo(MessageManager messageManager){
    	
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        if(this.agentInfo.getTime()%15==0)
        	this.goingback=false;
        this.fireclustering.updateInfo(messageManager);
        if(this.agentInfo.getTime()>=3 && !this.agentInfo.getPosition().equals(this.lastposition))
        {	
        	this.CalcWalkDirection(lastposition, this.agentInfo.getPosition());
        }
        this.CalcWater();
        this.CalcRefillTarget();
        this.pathPlanning.updateInfo(messageManager); 
        this.StuckFlag=this.isStuck();
        this.shouldIgo=this.Surrounded();
        this.AddCross();
        this.myclusterIndex=this.fireclustering.getClusterIndex(this.target);
        this.lastposition=this.agentInfo.getPosition();
        return this;
    }

    @Override
    public ExtAction setTarget(EntityID target) {
        this.target = null;
        if(target != null) {
            StandardEntity entity = this.worldInfo.getEntity(target);
            if(entity instanceof Building) {
                this.target = target;
            }
            else
            {
            	this.target=target;
            }
        }
        return this;
    }

    @Override
    public ExtAction calc() {
        this.result = null;
        FireBrigade agent = (FireBrigade)this.agentInfo.me();
        this.refillFlag = this.needRefill(agent, this.refillFlag);
        if(this.refillFlag) {
            this.result = this.calcRefill(agent, this.pathPlanning, this.target);
            if(this.result != null) {
            	this.updateWater();
                return this;
            }
        }

        if(this.needRest(agent)) {
            this.result = this.calcRefugeAction(agent, this.pathPlanning, this.target, false);
            if(this.result != null) {
            	this.updateWater();
                return this;
            }
        }

        if(this.target == null) {
        	this.updateWater();
        	this.result = null;
            return this;
        }
        this.updateWater();
        this.result = this.calcExtinguish(agent, this.pathPlanning, this.target);
        return this;
    }

    private Action calcExtinguish(FireBrigade agent, PathPlanning pathPlanning, EntityID target) {
    		EntityID agentPosition = agent.getPosition();
            StandardEntity positionEntity = Objects.requireNonNull(this.worldInfo.getPosition(agent));
            if(this.Mode==1)
            {
            	 if (StandardEntityURN.REFUGE == positionEntity.getStandardURN())
                 {
                     Action action = this.getMoveAction(pathPlanning, agentPosition, target);
                     if (action != null)
                     {
                         return action;
                     }
                 }
                 List<StandardEntity> neighbourBuilding = new ArrayList<>();
                 List<StandardEntity> NoseeneighbourBuilding = new ArrayList<>();
                 StandardEntity entity = this.worldInfo.getEntity(target);
                 if (entity instanceof Building)
                 {
                     if (this.worldInfo.getDistance(positionEntity, entity) < this.maxExtinguishDistance
                    		 && this.worldInfo.getChanged().getChangedEntities().contains(entity.getID()))
                     {
                         neighbourBuilding.add(entity);
                     }
                     if(this.worldInfo.getDistance(positionEntity, entity) < this.maxExtinguishDistance)
                    	 NoseeneighbourBuilding.add(entity);
                 }
                 if(this.StuckFlag&&NoseeneighbourBuilding.size()>0)
                	 return new ActionExtinguish(NoseeneighbourBuilding.get(0).getID(), this.maxExtinguishPower);
                 if (neighbourBuilding.size() > 0)
                 {
                     neighbourBuilding.sort(new DistanceSorter(this.worldInfo, agent));
                     return new ActionExtinguish(neighbourBuilding.get(0).getID(), this.maxExtinguishPower);
                 }
                 EntityID movetarget=this.SearchRightRoad(target);
                 return this.getMoveAction(pathPlanning, agentPosition, movetarget);
            }
            else
            {
            	if(StandardEntityURN.REFUGE == positionEntity.getStandardURN()) {
                    Action action = this.getMoveAction(pathPlanning, agentPosition, target);
                    if(action != null) {
                        return action;
                    }
                }
                List<StandardEntity> neighbourBuilding = new ArrayList<>();
                List<StandardEntity> NoseeneighbourBuilding = new ArrayList<>();
                StandardEntity entity = this.worldInfo.getEntity(target);
                if(entity instanceof Building) 
                {
                    if (this.worldInfo.getDistance(positionEntity, entity) < this.maxExtinguishDistance && this.worldInfo.getChanged().getChangedEntities().contains(entity.getID())) {
                        neighbourBuilding.add(entity);
                    }
                    if(this.worldInfo.getDistance(positionEntity, entity) < this.maxExtinguishDistance)
                   	 NoseeneighbourBuilding.add(entity);
                }
                else
                {	
                	return this.getMoveAction(pathPlanning, agentPosition, target);
                }
                
                if(this.StuckFlag&&NoseeneighbourBuilding.size()>0)
               	 return new ActionExtinguish(NoseeneighbourBuilding.get(0).getID(), this.maxExtinguishPower);
                if(neighbourBuilding.size() > 0)
                {
                    neighbourBuilding.sort(new DistanceSorter(this.worldInfo, agent));
                    if(this.StuckFlag)
                    {	
                    	return new ActionExtinguish(neighbourBuilding.get(0).getID(), this.ExtinguishPower);
                    }
                    if(this.shouldIgo&&this.goingback==false)//&&this.timedelay==8
                    {	
                    	this.goingout=true;
                    	this.CrossSort();
                    	this.DeleteCross();
                    	return this.getMoveAction(pathPlanning, agentPosition, this.cangocross.get(0).getID());
                    }
                    else if(this.StepOnFire())
                    {return this.getMoveAction(pathPlanning, agentPosition, this.GoOutofFire());}
                    else
                    {return new ActionExtinguish(neighbourBuilding.get(0).getID(), this.ExtinguishPower);}
                }
               if(this.goingout)
                	this.goingback=true;
               EntityID movetarget=this.SearchRightRoad(target);
               return this.getMoveAction(pathPlanning, agentPosition, movetarget);
            }
    }

    private Action getMoveAction(PathPlanning pathPlanning, EntityID from, EntityID target) {
        pathPlanning.setFrom(from);
        pathPlanning.setDestination(target);
        List<EntityID> path = pathPlanning.calc().getResult();
        if(path != null && path.size() > 0) {
            StandardEntity entity = this.worldInfo.getEntity(path.get(path.size() - 1));
            if(entity instanceof Building) {
                if(entity.getStandardURN() != StandardEntityURN.REFUGE) {
                    path.remove(path.size() - 1);
                }
            }
            return new ActionMove(path);
        }
        return null;
    }

    private boolean needRefill(FireBrigade agent, boolean refillFlag) {
        if(refillFlag) {
            StandardEntityURN positionURN = Objects.requireNonNull(this.worldInfo.getPosition(agent)).getStandardURN();
            return !(positionURN == REFUGE || positionURN == HYDRANT) || agent.getWater() < this.refillCompleted;
        }
        return  agent.getWater() <= this.refillRequest;
    }

    private boolean needRest(Human agent) {
        int hp = agent.getHP();
        int damage = agent.getDamage();
        if(hp == 0 || damage == 0) {
            return false;
        }
        int activeTime = (hp / damage) + ((hp % damage) != 0 ? 1 : 0);
        if(this.kernelTime == -1) {
            try {
                this.kernelTime = this.scenarioInfo.getKernelTimesteps();
            }catch (NoSuchConfigOptionException e) {
                this.kernelTime = -1;
            }
        }
        return damage >= this.thresholdRest || (activeTime + this.agentInfo.getTime()) < this.kernelTime;
    }
    
    private Action calcRefill(FireBrigade agent, PathPlanning pathPlanning, EntityID target)
    {
    	StandardEntityURN positionURN = Objects.requireNonNull(this.worldInfo.getPosition(agent)).getStandardURN();
        if(positionURN == REFUGE) {
            return new ActionRefill();
        }
        else if(positionURN == HYDRANT && this.agentInfo.getPosition().equals(this.refilltarget))
        {
        	return new ActionRefill();
        }
        EntityID dest=this.worldInfo.getEntity(this.refilltarget).getID();
        return this.getMoveAction(pathPlanning, this.agentInfo.getPosition(), dest);
    }
  
    private Action calcRefugeAction(Human human, PathPlanning pathPlanning, EntityID target, boolean isRefill)
    {
        return this.calcSupplyAction(
                human,
                pathPlanning,
                this.worldInfo.getEntityIDsOfType(StandardEntityURN.REFUGE),
                target,
                isRefill
        );
    }

    private Action calcSupplyAction(Human human, PathPlanning pathPlanning, Collection<EntityID> supplyPositions, EntityID target, boolean isRefill) {
        EntityID position = human.getPosition();
        int size = supplyPositions.size();
        if(supplyPositions.contains(position)) {
            return isRefill ? new ActionRefill() : new ActionRest();
        }
        List<EntityID> firstResult = null;
        while(supplyPositions.size() > 0) {
            pathPlanning.setFrom(position);
            pathPlanning.setDestination(supplyPositions);
            List<EntityID> path = pathPlanning.calc().getResult();
            if (path != null && path.size() > 0) {
                if (firstResult == null) {
                    firstResult = new ArrayList<>(path);
                    if(target == null) {
                        break;
                    }
                }
                EntityID supplyPositionID = path.get(path.size() - 1);
                pathPlanning.setFrom(supplyPositionID);
                pathPlanning.setDestination(target);
                List<EntityID> fromRefugeToTarget = pathPlanning.calc().getResult();
                if (fromRefugeToTarget != null && fromRefugeToTarget.size() > 0) {
                    return new ActionMove(path);
                }
                supplyPositions.remove(supplyPositionID);
                //remove failed
                if (size == supplyPositions.size()) {
                    break;
                }
                size = supplyPositions.size();
            } else {
                break;
            }
        }
     
        return firstResult != null ? new ActionMove(firstResult) : null;
    }
    
    private class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;

        DistanceSorter(WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
        }

        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = this.worldInfo.getDistance(this.reference, a);
            int d2 = this.worldInfo.getDistance(this.reference, b);
            return d1 - d2;
        }
    }
    
    private class CrossSorter implements Comparator<StandardEntity>
    {
    	private StandardEntity reference;
    	private WorldInfo worldInfo;
    	private Pair<Integer,Integer> direction;
    	private Pair<Integer,Integer> Walkdirection;
    	CrossSorter(WorldInfo wi, StandardEntity reference,Pair<Integer,Integer> dire,Pair<Integer,Integer> Walk)
    	{
    		this.reference = reference;
            this.worldInfo = wi;
            this.direction=dire;
            this.Walkdirection=Walk;
    	}
    	
    	public int compare(StandardEntity a,StandardEntity b)
    	{
    		FireBrigade me=(FireBrigade) reference;
    		List<StandardEntity> firebrigades=new ArrayList<>();
    		for(StandardEntity entity:this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE))
    		{
    			firebrigades.add(entity);
    		}
    		
    		int firebrigadenumA=0;
        	int firebrigadenumB=0;
        	
        	int AgentCostofA=0;int DistanceCostofA=0;int ClosestCostofA=500000;int DirectionCostofA=0;
        	int AgentCostofB=0;int DistanceCostofB=0;int ClosestCostofB=500000;int DirectionCostofB=0;
        	
        	int Acost=0;
        	int Bcost=0;
        	
        	double Length=Math.hypot(this.direction.first(), this.direction.second());
        	double DireX=this.direction.first()/Length;
        	double DireY=this.direction.second()/Length;
        	
        	int VecWalkX=0;//=this.Walkdirection.first()/10;
    		int VecWalkY=0;//=this.Walkdirection.second()/10;
    		if(this.Walkdirection!=null)
    		{
    			 VecWalkX=this.Walkdirection.first()/10;
        		 VecWalkY=this.Walkdirection.second()/10;
    		}
    		double VecAx=0,VecAy=0,VecBx=0,VecBy=0;
        	int price=500000;int disthreshold=15000;int agentthreshold=2;
        	for(StandardEntity entity:this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE))
            {
               if(this.worldInfo.getDistance(entity.getID(), a.getID())<disthreshold )
             		firebrigadenumA++;
               if(this.worldInfo.getDistance(entity.getID(), b.getID())<disthreshold )
             		firebrigadenumB++;
            }
        	
        	AgentCostofA=firebrigadenumA*firebrigadenumA*price;
        	AgentCostofB=firebrigadenumB*firebrigadenumB*price;
        	
        	DistanceCostofA=this.worldInfo.getDistance(me.getID(), a.getID());
        	DistanceCostofB=this.worldInfo.getDistance(me.getID(), b.getID());
        	
        	firebrigades.sort(new DistanceSorter(this.worldInfo,a));
        	for(int i=0;i<agentthreshold;i++)
        	{
        		if(firebrigades.get(i).equals(this.reference))
        		{
        			ClosestCostofA=0;
        			break;
        		}
        	}
        	firebrigades.sort(new DistanceSorter(this.worldInfo,b));
        	for(int i=0;i<agentthreshold;i++)
        	{
        		if(firebrigades.get(i).equals(this.reference))
        		{
        			ClosestCostofB=0;
        			break;
        		}
        	}
        	
        	if(a instanceof Road&&b instanceof Road)
        	{
        		Road roadA=(Road) a;
        		Road roadB=(Road) b;
        		
        		int myX=me.getX();int myY=me.getY();
        		int roadAx=roadA.getX();int roadAy=roadA.getY();
        		int roadBx=roadB.getX();int roadBy=roadB.getY();
        		
        		
        		VecAx=roadAx-myX;VecAy=roadAy-myY;
        		double LengthVecA=Math.hypot(VecAx, VecAy);
        		VecBx=roadBx-myX;VecBy=roadBy-myY;
        		double LengthVecB=Math.hypot(VecBx, VecBy);
        		
        		VecAx=VecAx/LengthVecA;VecAy=VecAy/LengthVecA;
        		VecBx=VecBx/LengthVecB;VecBy=VecBy/LengthVecB;
        		double cosA=(VecAx*DireX+VecAy*DireY)/(Math.hypot(VecAx, VecAy)*Math.hypot(DireX, DireY));
        		double cosB=(VecBx*DireX+VecBy*DireY)/(Math.hypot(VecBx, VecBy)*Math.hypot(DireX, DireY));
        		
        		if(cosA>=0&&VecWalkX*VecAx+VecWalkY*VecAy>0)
        			DirectionCostofA=(int)(DistanceCostofA*(1-cosA));
        		else
        			DirectionCostofA=500000;
        		if(cosB>=0&&VecWalkX*VecBx+VecWalkY*VecBy>0)
        			DirectionCostofB=(int)(DistanceCostofB*(1-cosB));
        		else
        			DirectionCostofB=500000;
        	}
        	Acost=AgentCostofA+ClosestCostofA+DirectionCostofA;
        	Bcost=AgentCostofB+ClosestCostofB+DirectionCostofB;
        	return Acost-Bcost;
    	}
    }
    
    private boolean isStuck()
    {
    	Human agent = (Human) agentInfo.me();
        int agentX = agent.getX();
        int agentY = agent.getY();
        StandardEntity positionEntity = worldInfo.getPosition(agent);
        if (positionEntity instanceof Road)
        {
            Road road = (Road) positionEntity;
            if (road.isBlockadesDefined() && road.getBlockades().size() > 0)
            {
                for (Blockade blockade : worldInfo.getBlockades(road))
                {
                    if (blockade == null || !blockade.isApexesDefined())
                    { continue; }

                    if (this.isInside(agentX, agentY, blockade.getApexes()))
                    { return true; }
                }
            }
        }
        return false;
    }
    private boolean isInside(double pX, double pY, int[] apex)
    {
        Point2D p = new Point2D(pX, pY);
        Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
        Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
        double theta = this.getAngle(v1, v2);

        for (int i = 0; i < apex.length - 2; i += 2)
        {
            v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
            v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
            theta += this.getAngle(v1, v2);
        }
        return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
    }
    private double getAngle(Vector2D v1, Vector2D v2)
    {
        double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
        double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
        if (flag > 0)
        {
            return angle;
        }
        if (flag < 0)
        {
            return -1 * angle;
        }
        return 0.0D;
    }    
    
    private void CalcWater()
    {
    	if(this.worldInfo.getEntity(target) instanceof Building)
    	{
    		Building thisbuilding=(Building)this.worldInfo.getEntity(target);
    		int Area=0;//
    		if(thisbuilding.isTotalAreaDefined())
    			Area=thisbuilding.getTotalArea();
    		int Temp=0;//
    		if(thisbuilding.isTemperatureDefined())
    			Temp=thisbuilding.getTemperature();
    		int WaterAmount=Area*AreaCoe+Temp*TempCoe;
    		if(WaterAmount>=this.maxExtinguishPower)
    			WaterAmount=this.maxExtinguishPower;
    		else if(WaterAmount<this.maxExtinguishPower*0.1)
    			WaterAmount=(int)(this.maxExtinguishPower*0.1);
    		else if(WaterAmount<=this.maxExtinguishPower/10)
    			WaterAmount=this.maxExtinguishPower/10;
    		else if(WaterAmount>=this.agentInfo.getWater())
    			WaterAmount=this.agentInfo.getWater()-100;
    		this.ExtinguishPower=WaterAmount;
    		this.refillRequest=1000;
    		
    	}
    }
    
    private void CalcRefillTarget()
    {
    	Collection<StandardEntity> Agents=this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE);
    	Collection<StandardEntity> Targets=this.worldInfo.getEntitiesOfType(HYDRANT,REFUGE);
    	List<StandardEntity> RefillTargets=new ArrayList<>();
    	List<StandardEntity> FireAgents=new ArrayList<>();
    	for (StandardEntity s:Targets)
    	{
    		RefillTargets.add(s);
    	}
    	for(StandardEntity s:Agents)
    	{
    		FireAgents.add(s);
    	}
    	//RefillTargets.sort(new RefillSorter(this.worldInfo, this.agentInfo.me(),this.RefugeSpeed,this.HydrantSpeed,this.Movedistance,this.Tankamount,this.pathPlanning));
    	RefillTargets.sort(new DistanceSorter(this.worldInfo,this.agentInfo.me()));
    	for(StandardEntity s:RefillTargets)
    	{
    		FireAgents.sort(new DistanceSorter(this.worldInfo, s));
    		if(s instanceof Refuge)
    		{
    			this.refilltarget=s.getID();
    			break;
    		}
    		if(FireAgents.get(0).equals(this.agentInfo.me())&&!this.StuckFlag&&this.isClosest(s.getID())&&!this.isOccupied(s.getID()))
    		{	
    			this.refilltarget=s.getID();
    			break;	
    		}
    	}	
    }
    
    private boolean StepOnFire()
    {
    	EntityID myposition=this.agentInfo.getPosition();
    	StandardEntity entity=this.worldInfo.getEntity(myposition);
    	StandardEntityURN urn=entity.getStandardURN();
    	switch(urn)
    	{
    	case BUILDING:
    		Building building=(Building) entity;
    		if(building.isOnFire())
    			return true;
    		break;
    	default:
    		break;
    	}
    	return false;
    }
    
    private EntityID GoOutofFire()
    {
    	EntityID result=null;
    	EntityID myposition=this.agentInfo.getPosition();
    	List<StandardEntity> neighbors=new ArrayList<>();
    	for(EntityID entityid:this.worldInfo.getObjectIDsInRange(myposition, 100000))
    	{
    		neighbors.add(this.worldInfo.getEntity(entityid));
    	}
    	neighbors.sort(new DistanceSorter(this.worldInfo, this.agentInfo.me()));
    	for(StandardEntity s:neighbors)
    	{
    		if(!s.getStandardURN().equals(StandardEntityURN.BUILDING))
    		{
    			result=s.getID();
    			break;
    		}
    	}
    	return result;
    }
    
    private void FindCross()
    {
    	int number=0;
    	for(StandardEntity entity:this.worldInfo.getEntitiesOfType(StandardEntityURN.ROAD))
    	{
    		Road road=(Road) entity;
    		number=road.getNeighbours().size();
    		for(EntityID entityid:road.getNeighbours())
    		{
    			if(!(this.worldInfo.getEntity(entityid) instanceof Road))
    			{number--;}
    			else
    			{
    				Road neighborroad=(Road) this.worldInfo.getEntity(entityid);
    				for(EntityID s:neighborroad.getNeighbours())
    				{
    					if(!(this.worldInfo.getEntity(s) instanceof Road))
    					{number--;}
    				}
    			}
    		}
    		if(number>=3&&!this.cross.contains(entity))
    		{cross.add(entity);
    		}
    	}
    }
    
    private void CrossSort()
    {
    	if(this.myclusterIndex==-1||this.myclusterIndex>=this.fireclustering.GetAllDirection().size())
    	{
    		this.cangocross.sort(new DistanceSorter(this.worldInfo,this.agentInfo.me()));
    	}
    	else
    	{
    		this.cangocross.sort(new CrossSorter(this.worldInfo,this.agentInfo.me(),this.fireclustering.GetDirection(this.myclusterIndex),this.Walkdire));
    	}
    	this.removedcross.sort(new DistanceSorter(this.worldInfo,this.agentInfo.me()));
    }
    
    private boolean Surrounded()
    {
    	boolean surrounded=false;
    	double firebuildings=0;
    	double notfirebuildings=0;
    	double extinguishbuildings=0;
    	for(EntityID entityid:this.worldInfo.getChanged().getChangedEntities())
    	{
    		if(this.worldInfo.getEntity(entityid) instanceof Building&&!(this.worldInfo.getEntity(entityid) instanceof Refuge))
    		{
    			Building building=(Building)this.worldInfo.getEntity(entityid);
    			if(building.isOnFire()||building.getFieryness()==8)
    			{
    				firebuildings++;
    			}
    			else if(building.isFierynessDefined()&&building.getFieryness()>=4&&building.getFieryness()<=7)
    				extinguishbuildings++;
    			else
    				notfirebuildings++;
    		}
    	}
    	boolean around=false;
    	double myX=this.agentInfo.getX();double myY=this.agentInfo.getY();
    	for(EntityID entityid:this.worldInfo.getChanged().getChangedEntities())
    	{
    		if(this.worldInfo.getEntity(entityid) instanceof Building&&!(this.worldInfo.getEntity(entityid) instanceof Refuge))
    		{	Building building1=(Building)this.worldInfo.getEntity(entityid);
    			double building1X=building1.getX();double building1Y=building1.getY();
    			Pair<Double,Double> Vec1=this.Unity(myX, myY, building1X, building1Y);
    			for(EntityID entityid2:this.worldInfo.getChanged().getChangedEntities())
        		{
    				if(this.worldInfo.getEntity(entityid2) instanceof Building&&!(this.worldInfo.getEntity(entityid2) instanceof Refuge))
    				{
    					Building building2=(Building)this.worldInfo.getEntity(entityid2);
    					double building2X=building2.getX();double building2Y=building2.getY();
    					Pair<Double,Double> Vec2=this.Unity(myX, myY, building2X, building2Y);
    					if(this.PointMul(Vec1, Vec2)<=0)
    					{	
    						around=true;
    						break;
    					}
    				}
        		}
    		}
    		if(around=true)
    			break;
    	}
    	//if I was surrounded by fires,I should go to somewhere else to ensure that 
    	//I can always observed the outer layer of fires
    	if(firebuildings>2*notfirebuildings+extinguishbuildings*0.5)//&&around==true
    	{
    		surrounded=true;
    	}
       	return surrounded;
    }
    
    private void DeleteCross()
    {
    	int index=this.cangocross.indexOf(this.worldInfo.getEntity(this.agentInfo.getPosition()));
    	if(index!=-1)
    	{
    		this.removedcross.add(this.cangocross.get(index));
    		this.cangocross.remove(index);
    	}
    }
    
    private void AddCross()
    {
    	int firebuildings=0;
    	int notfirebuildings=0;
    	if(this.removedcross.size()>0)
    	{
    		for(EntityID entityid:this.worldInfo.getObjectIDsInRange(this.removedcross.get(0),30000))
	    	{
	    		if(this.worldInfo.getEntity(entityid) instanceof Building)
				{
					Building building=(Building)this.worldInfo.getEntity(entityid);
					if(building.isFierynessDefined())
					{
						if(building.isOnFire()||building.getFieryness()==8)
							firebuildings++;
						else
							notfirebuildings++;
					}
				}
	    	}
	    	if(firebuildings<notfirebuildings)
			{
	    		
	    		this.cangocross.add(this.removedcross.get(0));
				this.removedcross.remove(0);
	    		
			}
    	}
    }
    
    private Pair<Double,Double> Unity(double x1,double y1,double x2,double y2)
    {
    	double Length=(int)Math.hypot(x1-x2, y1-y2);
    	double dx=x2-x1;
    	double dy=y2-y1;
    	double resultx=dx/Length;
    	double resulty=dy/Length;
    	return new Pair<>(resultx,resulty);
    }
    
    private double PointMul(Pair<Double,Double> first,Pair<Double,Double> second)
    {
    	double result=first.first()*second.first()+first.second()*second.second();
    	return result;
    }

    /* store all fire brigade water information
     * In order to make the the information correct
     * you must send own fire brigade info to others 
     */
    private void updateWater()
    {
    	this.firebrigadeWater.clear();
    	for(StandardEntity entity:this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE))
    	{
    		FireBrigade fireagent=(FireBrigade) entity;
    		this.firebrigadeWater.put(entity.getID(), fireagent.getWater());
    	}
    }
    /*
     * judge whether a FireBrigade is refilling
     */
    private boolean isRefilling(EntityID entityID)
    {
    	if(this.firebrigadeWater.get(entityID)!=null)
	    {
    		FireBrigade fireagent=(FireBrigade)this.worldInfo.getEntity(entityID);
	    	int lastWater=this.firebrigadeWater.get(entityID);
	    	int nowWater=fireagent.getWater();
	    	if(nowWater-lastWater>0)
	    	{
	    		return true;
	    	}
    	}
    	return false;
    }
    /*
     * This function was used to judge whether a certain hydrant has already been occupied
     */
    private boolean isOccupied(EntityID hydrantID)
    {
    	EntityID myID=this.agentInfo.getID();
    	for(StandardEntity agents:this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE))
    	{
    		if(agents instanceof FireBrigade)
    		{
    			FireBrigade firebrigade=(FireBrigade) agents;
    			if(firebrigade.getPosition().equals(hydrantID)&&this.isRefilling(firebrigade.getID())&&!firebrigade.getID().equals(myID))
    			{
    				return true;
    			}
    		}
    	}
    	return false;
    }
    
    /*
     * This function was used to judge whether I am the closest to a certain refilling target
     */
    private boolean isClosest(EntityID targetID)
    {
    	EntityID myid=this.agentInfo.getID();
    	int mydistance=this.worldInfo.getDistance(myid, targetID);
    	for(StandardEntity agents:this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE))
    	{
    		if(this.worldInfo.getDistance(agents.getID(), targetID)<mydistance&&!agents.getID().equals(myid)&&!this.agentInfo.getPosition().equals(targetID))
    		{
    			return false;
    		}
    	}
    	return true;
    }

    private EntityID SearchRightRoad(EntityID target)
    {
    	List<StandardEntity> roads=new ArrayList<>();
    	for(StandardEntity entity:this.worldInfo.getEntitiesOfType(StandardEntityURN.ROAD)) //,StandardEntityURN.BUILDING))
    	{
    		//if(entity instanceof Building)
    		//{
    		//	Building building=(Building) entity;
    		//	if(!building.isOnFire())
    		//		roads.add(entity);
    		//}
    		//else
    		//{
    			roads.add(entity);
    		//}
    	}
    	roads.removeAll(this.Entrances);
    	StandardEntity standRoad=this.worldInfo.getEntity(this.agentInfo.getPosition());
    	if(standRoad instanceof Road && !this.worldInfo.getChanged().getChangedEntities().contains(target))
    	{
    		if(this.WrongPosition.get(target)==null)
        	{
        		List<StandardEntity> list=new ArrayList<>();
        		list.add(standRoad);
        		this.WrongPosition.put(target, list);
        	}
        	else
        	{
        		List<StandardEntity> list=this.WrongPosition.get(target);
        		if(!list.contains(standRoad))
        			list.add(standRoad);
        		this.WrongPosition.put(target, list);
        	}
    	}
    	if(this.WrongPosition.get(target)!=null)
    	{
    		roads.removeAll(this.WrongPosition.get(target));
    	}
    	roads.sort(new DistanceSorter(this.worldInfo,this.worldInfo.getEntity(target)));
    	return roads.get(0).getID();
    }
    private void CalcWalkDirection(EntityID last,EntityID now)
    {
    	if(this.worldInfo.getEntity(last) instanceof Road && this.worldInfo.getEntity(now) instanceof Road)
    	{
    		Road lastroad=(Road) this.worldInfo.getEntity(last);
    		Road nowroad=(Road) this.worldInfo.getEntity(now);
    		int direx=nowroad.getX()-lastroad.getX();
    		int direy=nowroad.getY()-lastroad.getY();
    		this.Walkdire=new Pair<>(direx,direy);
    	}
    }
}



