package SEU.module.algorithm;

import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.DynamicClustering;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

public class SEUFireClustering extends DynamicClustering
{
    
    private int groupingDistance;
    List<List<StandardEntity>> clusterList = new LinkedList<>();
    List<List<StandardEntity>> clusteredgeList = new LinkedList<>();
    
    List<Pair<Integer,Integer>> clusterCenter=new ArrayList<>();
    List<List<Pair<Integer,Integer>>> HistoryCenter=new LinkedList<>(); 
    
    List<Pair<Integer,Integer>> direction=new ArrayList<>();
    
    int testID=516746579;
    
    public SEUFireClustering(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData)
    {
        super(ai, wi, si, moduleManager, developData);
        double groupdis=(this.worldInfo.getBounds().getHeight()+this.worldInfo.getBounds().getWidth())/20.0;
        this.groupingDistance = developData.getInteger("SEU.module.algorithm.SEUFireClustering.groupingDistance", (int)groupdis);
    }

    /**
     * calculation phase; update cluster
     * @return own instance for method chaining
     */
    @Override
    public Clustering calc()
    {
    	 for (StandardEntity changedEntity : worldInfo.getEntitiesOfType(StandardEntityURN.BUILDING))
	     {
	                Building changedBuilding = (Building)changedEntity;
	                if (this.getClusterIndex(changedEntity) < 0)
	                { // cE is not contained in cluster
	                    if (isBurning(changedBuilding))
	                    { // cE is burning building
	                    	//System.out.println("find burning building."+changedBuilding.getID());
	                        ArrayList<EntityID> hostClusterPropertyEntityIDs = new ArrayList<>();
							//System.out.println("changedBuilding:"+ changedBuilding.getID());

	                        // search host cluster
	                        for (List<StandardEntity> cluster : this.clusterList)
	                        {
	                            for (StandardEntity entity : cluster)
	                            {
	                                if (worldInfo.getDistance(entity, changedBuilding) <= groupingDistance)
	                                {
	                                    hostClusterPropertyEntityIDs.add(entity.getID());
	                                    break;
	                                }
	                            }
	                        }

	                        if (hostClusterPropertyEntityIDs.size() == 0)
	                        { // there is not host cluster : form new cluster
	                            List<StandardEntity> cluster = new ArrayList<>();
	                            clusterList.add(cluster);
	                            cluster.add(changedBuilding);
	                            HistoryCenter.add(new LinkedList<>());
	                        }
	                        else if (hostClusterPropertyEntityIDs.size() == 1)
	                        { // there is one host cluster : add building to the cluster
	                            int hostIndex = this.getClusterIndex(hostClusterPropertyEntityIDs.get(0));
	                            clusterList.get(hostIndex).add(changedBuilding);
	                        }
	                        else
	                        { // there are multiple host clusters : add building to the cluster & combine clusters
	                            int hostIndex = this.getClusterIndex(hostClusterPropertyEntityIDs.get(0));
	                            List<StandardEntity> hostCluster = clusterList.get(hostIndex);
	                            if(!hostCluster.contains(changedBuilding))
	                            	hostCluster.add(changedBuilding);
	                            for (int index = 1; index < hostClusterPropertyEntityIDs.size(); index++)
	                            {
	                                int tergetClusterIndex = this.getClusterIndex(hostClusterPropertyEntityIDs.get(index));
	                                hostCluster.addAll(clusterList.get(tergetClusterIndex));
	                                //clusterList.remove(tergetClusterIndex);
	                                clusterList.get(tergetClusterIndex).clear();
	                            }
	                        }
	                    }
	                }
	                else
	                { // cE is contained in cluster
	                    if (!(isBurning(changedBuilding)))
	                    { // cE is not burning building
	                        int hostClusterIndex = this.getClusterIndex(changedBuilding);
	                        List<StandardEntity> hostCluster = clusterList.get(hostClusterIndex);
	                        hostCluster.remove(changedBuilding);
	                       
	                    }
	                }
	     }
        return this;
    }

    @Override
    public Clustering updateInfo(MessageManager messageManager)
    {
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() > 1)
        { return this; }
        this.calc(); // invoke calc()
        this.calcCenter();
        this.calcAllDire();
        this.calcEdges();
        
        this.debugStdOut("Cluster : " + clusterList.size());

        return this;
    }

    @Override
    public Clustering precompute(PrecomputeData precomputeData)
    {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() > 1) { return this; }
        return this;
    }

    @Override
    public Clustering resume(PrecomputeData precomputeData)
    {
        super.resume(precomputeData);
        if(this.getCountResume() > 1) { return this; }
        return this;
    }

    @Override
    public Clustering preparate()
    {
        super.preparate();
        if(this.getCountPreparate() > 1) { return this; }
        return this;
    }

    @Override
    public int getClusterNumber()
    {
        return clusterList.size();
    }

    @Override
    public int getClusterIndex(StandardEntity standardEntity)
    {
        for (int index = 0; index < clusterList.size(); index++)
        {
            if (clusterList.get(index).contains(standardEntity))
            { return index; }
        }
        return -1;
    }

    @Override
    public int getClusterIndex(EntityID entityID)
    {
        return getClusterIndex(worldInfo.getEntity(entityID));
    }

    @Override
    public Collection<StandardEntity> getClusterEntities(int i)
    {
        return clusterList.get(i);
    }

    @Override
    public Collection<EntityID> getClusterEntityIDs(int i)
    {
        ArrayList<EntityID> list = new ArrayList<>();
        for (StandardEntity entity : getClusterEntities(i))
        { list.add(entity.getID()); }
        return list;
    }


    /**
     * classify burning building
     * @param building target building
     * @return is building burning
     */
    private boolean isBurning(Building building)
    {
        if (building.isFierynessDefined())
        {
            switch (building.getFieryness())
            {
                case 1: case 2: case 3:
                    return true;
                default:
                    return false;
            }
        }
        return false;
    }
    
    /**
     * output text with class name to STDOUT when debug-mode.
     * @param text output text
     */
    private void debugStdOut(String text)
    {
        if (scenarioInfo.isDebugMode())
        { System.out.println("[" + this.getClass().getSimpleName() + "] " + text); }
    }

    private void calcCenter()
    {
    	this.clusterCenter.clear();
    	for(List<StandardEntity> cl:this.clusterList)
    	{   
    		int totalarea=0;
    		double x=0,y=0;
    		for(StandardEntity entity:cl)
    		{
    			if(entity instanceof Building)
    			{
    				Building building=(Building) entity;
    				totalarea+=building.getTotalArea();
    			}
    		}
    		for(StandardEntity entity:cl)
    		{
    			if(entity instanceof Building)
    			{
    				Building building=(Building) entity;
    				double property=(double)building.getTotalArea()/(double)totalarea;
    				x+=property*building.getX();
    				y+=property*building.getY();
    				
    			}
    		}
    		Pair<Integer, Integer>my=new Pair<>((int)x,(int)y);
    		this.clusterCenter.add(my);
    		int clusterIndex=clusterList.indexOf(cl);
    		this.HistoryCenter.get(clusterIndex).add(my);
    	}
    }
    private List<Pair<Integer,Integer>> calcAllDire()
    {
    	List<Pair<Integer,Integer>> result=new ArrayList<>();
    	Pair<Integer,Integer> m=new Pair<>(0,0);
    	Pair<Integer,Integer> n=new Pair<>(0,0);
    	
    	for(List<Pair<Integer,Integer>> list:this.HistoryCenter)
    	{   
    		if(list.size()>=6)
    		{
    			Pair<Integer, Integer> one=list.get(list.size()-1);
    			Pair<Integer, Integer> two=list.get(0);
    			n=this.calcDire(one, two);
    			if(!n.equals(m))
    			{
    				int index=this.HistoryCenter.indexOf(list);
    				for(int i=this.direction.size()-1;i<index;i++)
    				{
    					this.direction.add(m);
    				}
    				int lastx=this.direction.get(index).first();
    				int lasty=this.direction.get(index).second();
    				int lastx2=n.first();
    				int lasty2=n.second();
    				double first=lastx*0.875+0.125*lastx2;
    				double second=lasty*0.875+0.125*lasty2;
    				Pair <Integer,Integer>re=new Pair<>((int)first,(int)second);
    				this.direction.set(index, re);
    			}
    			result.add(n);
    		}
    		else
    		{
    			result.add(m);
    		}
    	}
    	return result;
    }
    private Pair<Integer,Integer> calcDire(Pair<Integer,Integer> one,Pair<Integer,Integer> two)
    {
    	int dx=one.first()-two.first();
    	int dy=one.second()-two.second();
    	Pair<Integer,Integer> m=new Pair<>(dx,dy);
    	return m;
    }
    private void calcEdges()
    {	this.clusteredgeList.clear();
    	for(List<StandardEntity> list:this.clusterList)
    	{
    		List<StandardEntity> cluster=new ArrayList<>();
    		Pair<Integer,Integer> center=this.GetCenter(this.clusterList.indexOf(list));
    		for(StandardEntity entity:list)
    		{
    			if(entity instanceof Building)
    			{
    				Building building=(Building) entity;
    				for(StandardEntity neighbor:this.worldInfo.getObjectsInRange(entity, this.groupingDistance/5))
    				{
    					if(neighbor instanceof Building)
    					{
    						Building neibuilding=(Building) neighbor;
    						int dx=building.getX()-center.first();
    						int dy=building.getY()-center.second();
    						int dx1=neibuilding.getX()-building.getX();
    						int dy1=neibuilding.getX()-building.getY();
    						if(dx*dx1+dy*dy1>=0&&!neibuilding.isOnFire()&&neibuilding.isFierynessDefined()&&neibuilding.getFieryness()!=8)
    						{
    							cluster.add(building);
    							break;
    						}
    					}
    				}
    			}
    		}	
    		this.clusteredgeList.add(cluster);
    	}
    }
    
    public Pair<Integer,Integer> GetDirection(int index)
    {return this.direction.get(index);}
    public Pair<Integer,Integer> GetCenter(int index)
    {return this.clusterCenter.get(index);}
    public List<Pair<Integer,Integer>> GetAllDirection()
    {return this.direction;}
    public List<Pair<Integer,Integer>> GetAllCenter()
    {return this.clusterCenter;}
    public List<StandardEntity> GetEdges(int index)
    {return this.clusteredgeList.get(index);}
    public List<List<StandardEntity>> GetAllEdges()
    {return this.clusteredgeList;}
    public List<Integer> GetEdgeIDvalues(int index)
    {
    	List<Integer> result=new ArrayList<>();
    	for(StandardEntity entity:this.GetEdges(index))
    	{
    		result.add(entity.getID().getValue());
    	}
    	return result;
    }
    public void PrintCluster()
    {
    	for(List<StandardEntity> list:this.clusterList)
    	{
    		System.out.print("cluster"+this.clusterList.indexOf(list)+": ");
    		for(StandardEntity entity:list)
    		{
    			System.out.print(entity.getID().getValue()+"\t");
    		}
    		System.out.println("");
    	}
    }
    public void PrintEdgeCluster()
    {
    	for(List<StandardEntity> list:this.clusteredgeList)
    	{
    		System.out.print("Edge"+this.clusteredgeList.indexOf(list)+": ");
    		for(StandardEntity entity:list)
    		{
    			System.out.print(entity.getID().getValue()+"\t");
    		}
    		System.out.println("");
    	}
    }
}





