package SEU.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandFire;
import adf.agent.communication.standard.bundle.centralized.MessageReport;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.FireTargetAllocator;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import SEU.module.algorithm.SEUFireClustering;

import java.util.*;

/**
 * @author lwj
 *
 */
public class SEUFireTargetAllocator extends FireTargetAllocator {

	private Clustering clustering;
    private Collection<EntityID> priorityBuildings;
    private Collection<EntityID> targetBuildings;

    private Map<EntityID, FireBrigadeInfo> agentInfoMap;
    //added
	private SEUFireClustering fireclustering;
    private int maxWater;
    private int maxPower;
    //added by lwj
    private Map<StandardEntity, StandardEntity> commandHistory;//used by store the command history and avoid to
    //give agents similar commands, only when the current target is far from the history target, center
    //will give the command to the agents

    public SEUFireTargetAllocator(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        this.commandHistory = new HashMap<>();
        this.priorityBuildings = new HashSet<>();
        this.targetBuildings = new HashSet<>();
        this.agentInfoMap = new HashMap<>();
        this.maxWater = si.getFireTankMaximum();
        this.maxPower = si.getFireExtinguishMaxSum();
        this.clustering = moduleManager.getModule("SEUFireTargetAllocator.Clustering", "adf.sample.module.algorithm.SampleFireClustering");
        this.fireclustering = moduleManager.getModule("SEUFireTargetAllocator.SEUFireClutering", "adf.sample.module.algorithm.SampleFireClustering");
    }

    @Override
    public FireTargetAllocator resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        for(EntityID id : this.worldInfo.getEntityIDsOfType(StandardEntityURN.FIRE_BRIGADE)) {
            this.agentInfoMap.put(id, new FireBrigadeInfo(id));
        }
        return this;
    }

    @Override
    public FireTargetAllocator preparate() {
        super.preparate();
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        for(EntityID id : this.worldInfo.getEntityIDsOfType(StandardEntityURN.FIRE_BRIGADE)) {
            this.agentInfoMap.put(id, new FireBrigadeInfo(id));
        }
        return this;
    }

    @Override
    public Map<EntityID, EntityID> getResult() {
        return this.convert(this.agentInfoMap);
    }

    @Override
    public FireTargetAllocator calc() {
        int currentTime = this.agentInfo.getTime();
        List<StandardEntity> agents = this.getActionAgents(this.agentInfoMap);
        Collection<EntityID> removes = new ArrayList<>();
        //codes below is added by lwj
        List<Integer> FireSize = new ArrayList<>();
        for(int i = 0;i < fireclustering.getClusterNumber();++i)
        {
        	FireSize.add(CalcFireClusterSize(fireclustering.getClusterEntities(i)));
        }
        List<Integer> distributeagents = CalcNumofFireBridge(FireSize, agents.size());
        if (distributeagents.size() != 0 ) {
		//下面的函数目的是给每个火势分区分配对应数量的火警,但是在火警目标的选取上较为简单
		//随机选择该分区边缘房屋中的一栋楼房,分配到该区的火警都会选择该楼房作为目标
			for(int i = 0;i < fireclustering.getClusterNumber();++i)
			{
				StandardEntity target = null;
				Random random = new Random(this.agentInfo.getID().getValue());
				int clusterSize = fireclustering.getClusterEntities(i).size();
				int flag = 0;
				if(clusterSize != 0)
					flag = random.nextInt(clusterSize);
				int temp = 0;
				for(StandardEntity randomSelectBuilding : fireclustering.getClusterEntities(i))
				{
					if (temp != flag) {
						++temp;
						continue;
					}
					int num = distributeagents.get(i);
					target = randomSelectBuilding;
					agents.sort(new DistanceSorter(this.worldInfo, target));
					for(int j = 0;j<num;++j)
					{
						StandardEntity result = agents.get(0);
						if (commandHistory.containsKey(result)) {
							StandardEntity historytarget = commandHistory.get(result);
							int distance = this.worldInfo.getDistance(historytarget, target);
							if(distance<100000)
								continue;
						}
						agents.remove(0);
						FireBrigadeInfo info = this.agentInfoMap.get(result.getID());
						if(info != null) {
							info.canNewAction = false;
							info.target = target.getID();
							info.commandTime = currentTime;
							this.commandHistory.put(result, target);
							this.agentInfoMap.put(result.getID(), info);
							System.out.println("in targetallocator " + result.getID() + " target is " + info.target);
							removes.add(target.getID());
						}
					}
					break;
				}
				
			}
			return this;
		}
        //原代码主要思路是:在所有可以执行新任务的agent中,中心会选择离目标building最近的一个agent,并将其目标设为该building
        for(EntityID target : this.priorityBuildings) {
            if(agents.size() > 0) {
                StandardEntity targetEntity = this.worldInfo.getEntity(target);
                if (targetEntity != null) {
                    agents.sort(new DistanceSorter(this.worldInfo, targetEntity));
                    StandardEntity result = agents.get(0);
                    agents.remove(0);
                    FireBrigadeInfo info = this.agentInfoMap.get(result.getID());
                    if(info != null) {
                        info.canNewAction = false;
                        info.target = target;
                        info.commandTime = currentTime;
                        this.agentInfoMap.put(result.getID(), info);
                        removes.add(target);
                    }
                }
            }
        }
        this.priorityBuildings.removeAll(removes);
        removes.clear();
        for(EntityID target : this.targetBuildings) {
            if(agents.size() > 0) {
                StandardEntity targetEntity = this.worldInfo.getEntity(target);
                if (targetEntity != null) {
                    agents.sort(new DistanceSorter(this.worldInfo, targetEntity));
                    StandardEntity result = agents.get(0);
                    agents.remove(0);
                    FireBrigadeInfo info = this.agentInfoMap.get(result.getID());
                    if(info != null) {
                        info.canNewAction = false;
                        info.target = target;
                        info.commandTime = currentTime;
                        this.agentInfoMap.put(result.getID(), info);
                        removes.add(target);
                    }
                }
            }
        }
        this.targetBuildings.removeAll(removes);
        return this;
    }

    @Override
    public FireTargetAllocator updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
		this.fireclustering.updateInfo(messageManager);//更新clustering的信息
        int currentTime = this.agentInfo.getTime();
        for(CommunicationMessage message : messageManager.getReceivedMessageList(MessageBuilding.class)) {
            MessageBuilding mb = (MessageBuilding)message;
            Building building = MessageUtil.reflectMessage(this.worldInfo, mb);
            if(building.isOnFire()) {
                this.targetBuildings.add(building.getID());
            } else {
                this.priorityBuildings.remove(mb.getBuildingID());
                this.targetBuildings.remove(mb.getBuildingID());
            }
        }
        for(CommunicationMessage message : messageManager.getReceivedMessageList(MessageFireBrigade.class)) {
            MessageFireBrigade mfb = (MessageFireBrigade) message;
            MessageUtil.reflectMessage(this.worldInfo, mfb);
            FireBrigadeInfo info = this.agentInfoMap.get(mfb.getAgentID());
            if(info == null) {
                info = new FireBrigadeInfo(mfb.getAgentID());
            }
            if(currentTime >= info.commandTime + 2) {
                this.agentInfoMap.put(mfb.getAgentID(), this.update(info, mfb));
            }
        }
        for(CommunicationMessage message : messageManager.getReceivedMessageList(CommandFire.class)) {
            CommandFire command = (CommandFire)message;
            if(command.getAction() == CommandFire.ACTION_EXTINGUISH && command.isBroadcast()) {
                this.priorityBuildings.add(command.getTargetID());
                this.targetBuildings.add(command.getTargetID());
            }
        }
        for(CommunicationMessage message : messageManager.getReceivedMessageList(MessageReport.class)) {
            MessageReport report = (MessageReport) message;
            FireBrigadeInfo info = this.agentInfoMap.get(report.getSenderID());
            if(info != null && report.isDone()) {
                info.canNewAction = true;
                this.priorityBuildings.remove(info.target);
                this.targetBuildings.remove(info.target);
                info.target = null;
                this.agentInfoMap.put(report.getSenderID(), info);
            }
        }
        return this;
    }

    private Map<EntityID, EntityID> convert(Map<EntityID, FireBrigadeInfo> infoMap) {
        Map<EntityID, EntityID> result = new HashMap<>();
        for(EntityID id : infoMap.keySet()) {
            FireBrigadeInfo info = infoMap.get(id);
            if(info != null && info.target != null) {
                result.put(id, info.target);
            }
        }
        return result;
    }

    private List<StandardEntity> getActionAgents(Map<EntityID, FireBrigadeInfo> infoMap) {
        List<StandardEntity> result = new ArrayList<>();
        //获得所有能进行新动作的agent
        for(StandardEntity entity : this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)) {
            FireBrigadeInfo info = infoMap.get(entity.getID());
            if(info != null && info.canNewAction && ((FireBrigade)entity).isPositionDefined()) {
                result.add(entity);
            }
        }
        return result;
    }

    private FireBrigadeInfo update(FireBrigadeInfo info, MessageFireBrigade message) {
        if(message.isBuriednessDefined() && message.getBuriedness() > 0) {
            info.canNewAction = false;
            if (info.target != null) {
                this.targetBuildings.add(info.target);
                info.target = null;
            }
            return info;
        }
        if(message.getAction() == MessageFireBrigade.ACTION_REST) {
            info.canNewAction = true;
            if (info.target != null) {
                this.targetBuildings.add(info.target);
                info.target = null;
            }
        } else if(message.getAction() == MessageFireBrigade.ACTION_REFILL) {
            info.canNewAction = (message.getWater() + this.maxPower >= this.maxWater);
            if (info.target != null) {
                this.targetBuildings.add(info.target);
                info.target = null;
            }
        } else if(message.getAction() == MessageFireBrigade.ACTION_MOVE) {
            if(message.getTargetID() != null) {
                StandardEntity entity = this.worldInfo.getEntity(message.getTargetID());
                if(entity != null && entity instanceof Area) {
                    if(info.target != null) {
                        StandardEntity targetEntity = this.worldInfo.getEntity(info.target);
                        if(targetEntity != null && targetEntity instanceof Area) {
                            if(message.getTargetID().getValue() == info.target.getValue()) {
                                info.canNewAction = false;
                            } else {
                                info.canNewAction = true;
                                this.targetBuildings.add(info.target);
                                info.target = null;
                            }
                        } else {
                            info.canNewAction = true;
                            info.target = null;
                        }
                    } else {
                        info.canNewAction = true;
                    }
                } else {
                    info.canNewAction = true;
                    if(info.target != null) {
                        this.targetBuildings.add(info.target);
                        info.target = null;
                    }
                }
            } else {
                info.canNewAction = true;
                if(info.target != null) {
                    this.targetBuildings.add(info.target);
                    info.target = null;
                }
            }
        } else if(message.getAction() == MessageFireBrigade.ACTION_EXTINGUISH) {
            info.canNewAction = true;
            info.target = null;
            this.priorityBuildings.remove(message.getTargetID());
            this.targetBuildings.remove(message.getTargetID());
        }
        return info;
    }

    private class FireBrigadeInfo {
        EntityID agentID;
        EntityID target;
        boolean canNewAction;
        int commandTime;

        FireBrigadeInfo(EntityID id) {
            agentID = id;
            target = null;
            canNewAction = true;
            commandTime = -1;
        }
    }

    private class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;

        DistanceSorter(WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
        }

        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = this.worldInfo.getDistance(this.reference, a);
            int d2 = this.worldInfo.getDistance(this.reference, b);
            return d1 - d2;
        }
    }
    
    
    /**
     * @param area
     * 代表一片着火的区域
     * @return
     * 返回值为一个最终评估值,用来反映一个火区着火情况的严重性,越大代表火势越严重
     */
    public int CalcFireClusterSize(Collection<StandardEntity> area){
    	int size = 0;
/*        for(StandardEntity standardEntity : area)
 *        {
 *            Building building = null;
 *            if(standardEntity instanceof Building)
 *                building = (Building)standardEntity;
 *            if (building != null && building.isTemperatureDefined() && building.getImportance()>5) {
 *
 *            }
 *        }*/
		size = area.size();
    	return size;
    }
    
    /**
     * @param FireSize
     * 是一个集合,包含每一个火区的火势评估值
     * @param NumofAllFireBridge
     * 是所有可以分配新任务的agent的数量
     * @return
     * 返回一个list,包含每个火区实际分配的agent的数量
     */
    public List<Integer> CalcNumofFireBridge(Collection<Integer> FireSize,int NumofAllFireBridge)
    {
    	List<Integer> result = new ArrayList<>();
    	if(FireSize.size() == 0 || NumofAllFireBridge == 0)
    	{
    		//System.out.println("FireSize:" + FireSize.size() + " NumofAllFireBridge:" + NumofAllFireBridge );
    		return result;
    	}
    	int sum = 0;//sum代表所有size之和,按照每个区size占sum中的比例来分配人员
    	for(int i : FireSize)
    	{
    		sum += i;
    	}
    	if (sum == 0) {
    		//System.out.println("sum == 0");
			return result;
		}
		int distributesum = 0;
    	for(int i : FireSize)
    	{
			int distributeNum = (int)((i / sum) * NumofAllFireBridge);
			distributesum += distributeNum;
			result.add(distributeNum);
    	}
    	int sizeofFireArea = result.size();
    	if (sizeofFireArea == 0) {
			return result;
		}
		//下面的目的是将剩余的agent平分到各个区去
		for(int i = 0; i < NumofAllFireBridge - distributesum; ++i)
		{
			result.set(i%sizeofFireArea, result.get(i%sizeofFireArea)+1);
		}
		//System.out.println("number of agents that be distributed is:" + NumofAllFireBridge);
		//int temp = 0;
		//for(int i : result)
		//if there is only one firecluster, we will not allocate all the fbs to the firecluster
		if(result.size() == 1)
			result.set(0, (Integer)(result.get(0)/2));
    	return result;
    }
}

