package SEU.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_CENTRE;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_OFFICE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;

import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.information.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import adf.debug.TestLogger;
import org.apache.log4j.Logger;

import adf.component.communication.CommunicationMessage;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.complex.Search;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

public class SEUSearchForFire extends Search {
	private PathPlanning pathPlanning;
	private Clustering clustering;
	private Set<EntityID> emptyBuildings;
	private Set<EntityID> burnBuildings;
	private EntityID result;
	private Collection<EntityID> unsearchedBuildingIDs;
	private List<EntityID> unsearchedRoadIDs;
	private Collection<EntityID> blockedBuildingIDs;
	private Logger logger;
	
	//Repeat times for searching a certain cluster
	private int threshold=5;
	private int SearchTime=threshold;
	private int SearchclusterIndex=0;
	
	/*------------------------------------
	 *Flags to judge the searching condition 
	 */
	private boolean ContainFire=false;
	private boolean Distributed=false;
	private boolean BurnOut=false;
	
	//------------------------------------

	public SEUSearchForFire(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
		super(ai, wi, si, moduleManager, developData);
		logger = TestLogger.getLogger(agentInfo.me());
		this.unsearchedBuildingIDs = new HashSet<>();
		this.unsearchedRoadIDs=new ArrayList<>();
		this.emptyBuildings=new HashSet<>();
		this.burnBuildings=new HashSet<>();
		this.blockedBuildingIDs=new HashSet<>();
		StandardEntityURN agentURN = ai.me().getStandardURN();
		if (agentURN == AMBULANCE_TEAM) {
			this.pathPlanning = moduleManager.getModule("SEUSearch.PathPlanning.Ambulance", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("SEUSearch.Clustering.Ambulance", "adf.sample.module.algorithm.SampleKMeans");
		} else if (agentURN == FIRE_BRIGADE) {
			this.pathPlanning = moduleManager.getModule("SEUSearch.PathPlanning.Fire", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("SEUSearch.Clustering.Fire", "adf.sample.module.algorithm.SampleKMeans");
		} else if (agentURN == POLICE_FORCE) {
			this.pathPlanning = moduleManager.getModule("SEUSearch.PathPlanning.Police", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("SEUSearch.Clustering.Police", "adf.sample.module.algorithm.SampleKMeans");
		}
		registerModule(this.clustering);
		registerModule(this.pathPlanning);
	}

	@Override
	public Search updateInfo(MessageManager messageManager) {
		logger.debug("Time:" + agentInfo.getTime());
		super.updateInfo(messageManager);
		this.reflectMessage(messageManager);
		if(this.agentInfo.me() instanceof FireBrigade)
		{
			super.updateInfo(messageManager);
			this.unsearchedRoadIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());
			this.JudgeFire();
			this.BurnOut=this.Burnout();
			if(this.ContainFire)
				this.SearchTime=threshold;
			if(BurnOut)
			{
				this.SearchTime=0;
				reset();
			}
			if(this.unsearchedRoadIDs.isEmpty()) {
	            this.reset();
	        }
	        this.unsearchedRoadIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());

	        if(this.unsearchedRoadIDs.isEmpty()) {
	            this.reset();
	            this.unsearchedRoadIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());
	        }
	        return this;
		}
		else
		{
		this.unsearchedBuildingIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());
		this.updateemptybuildings(); 
		this.updateemptybuildings();
		this.unsearchedBuildingIDs.removeAll(burnBuildings);
		if (!unsearchedBuildingIDs.isEmpty()) {
				for(EntityID building:unsearchedBuildingIDs)
				{
					if(isEntranceblock(building))
						this.blockedBuildingIDs.add(building);
				}
				this.unsearchedBuildingIDs.removeAll(blockedBuildingIDs);
			}

			if (this.unsearchedBuildingIDs.isEmpty()) {
				this.reset();
				this.unsearchedBuildingIDs.removeAll(this.worldInfo.getChanged().getChangedEntities());
			if(this.worldInfo.getEntity(this.agentInfo.getID()) instanceof AmbulanceTeam){
			this.unsearchedBuildingIDs.removeAll(emptyBuildings);
			this.unsearchedBuildingIDs.removeAll(burnBuildings);
			if (this.unsearchedBuildingIDs.isEmpty()){
				 int clusterIndex = this.clustering.getClusterIndex(this.agentInfo.getID());
				 for (int i = 0; i < this.clustering.getClusterNumber(); ++i) {
					 if(clusterIndex!=i)
				unsearchedBuildingIDs.addAll(this.clustering.getClusterEntityIDs(i));
		     		}
				 }
			}
			}
			return this;
		}	
	}

	@Override
	public Search calc() {
		if(this.agentInfo.me() instanceof FireBrigade)
		{
			this.result = null;
			if (unsearchedRoadIDs.isEmpty())
				return this;
			logger.debug("unsearchedRoadIDs:" + unsearchedRoadIDs);
			this.pathPlanning.setFrom(this.agentInfo.getPosition());
			this.pathPlanning.setDestination(this.unsearchedRoadIDs);
			List<EntityID> path = this.pathPlanning.calc().getResult();
			logger.debug("best path is:"+path);
			if (path != null) {
				this.result = path.get(path.size() - 1);
			}
			logger.debug("choosed:"+result);
			return this;
		}
		else
		{
			this.result = null;
			if (unsearchedBuildingIDs.isEmpty())
				return this;

			logger.debug("unsearchedBuildingIDs:" + unsearchedBuildingIDs);
			this.pathPlanning.setFrom(this.agentInfo.getPosition());
			this.pathPlanning.setDestination(this.unsearchedBuildingIDs);
			List<EntityID> path = this.pathPlanning.calc().getResult();
			logger.debug("best path is:"+path);
			if (path != null && path.size() > 2) {
				this.result = path.get(path.size() - 2);
			}else if (path != null && path.size() > 0) {
				this.result = path.get(path.size() - 1);
			}
			logger.debug("choosed:"+result);
			return this;
		}
	}

	private boolean isEntranceblock(EntityID targetid )
	   {
	        StandardEntity entity = this.worldInfo.getEntity(targetid);
	        if(entity != null && entity instanceof Building) 
	        {	                    
	                        for(EntityID neighbor:((Building)entity).getNeighbours()) 
	                        {
	                            StandardEntity entity1 = this.worldInfo.getEntity(neighbor);
	                            if(entity1 != null && entity1 instanceof Road)
	                            {
	                                 Road road=(Road)entity1;
	                                 if(road.isBlockadesDefined()&&road.getBlockades().size()>0)
	                                 {
	                                   return true;
	                                 }
	                            }
	                        }
	                    }  
	        return false;
	   }
	
	  private void updateemptybuildings() {
          EntityID me=this.agentInfo.getID();
          StandardEntity entity = this.worldInfo.getEntity(me);
          if(entity instanceof AmbulanceTeam){
              Human human = (Human)entity;
              EntityID position = human.getPosition();
              if (this.worldInfo.getEntity(position) instanceof Building) {               	 
              this.emptyBuildings.add(position);
             	 for (EntityID id : this.worldInfo.getChanged().getChangedEntities()){
             		 StandardEntity entity1 = this.worldInfo.getEntity(id);
                      if(entity1 instanceof Human){
                     	 Human human1 = (Human)entity;
                          EntityID position1 = human.getPosition();
                          if(position==position1){
                         	  if (human1.isBuriednessDefined()) {
                                   this.emptyBuildings.remove(position);
                          }
                      }
             	 }
              }
          }
      }
 }
	 
	  private void updateburnbuildings() {
		  this.burnBuildings.addAll(this.worldInfo.getFireBuildingIDs());	  
		  Collection<EntityID>buildings=this.worldInfo.getEntityIDsOfType(BUILDING);		 
		  for(EntityID building: buildings){
			  if (((Building)this.worldInfo.getEntity(building)).getFieryness() == 8) {
				  this.burnBuildings.add(building) ;
			  }
		  }
		
	  }
	  
	private void reset() {
		this.JudgeFire();
		if(this.Distributed==false)
		{
			this.SearchclusterIndex=this.clustering.getClusterIndex(this.agentInfo.getID());
			this.Distributed=true;
		}
		if(this.agentInfo.me() instanceof FireBrigade)
		{
			if(this.ContainFire==true)
				this.SearchTime=threshold;
			else
				this.SearchTime--;
			
			if(this.SearchTime!=0)
			{
				this.unsearchedRoadIDs.clear();
		        Collection<StandardEntity> clusterEntities = null;
		        if(this.clustering != null) {
		            clusterEntities = this.clustering.getClusterEntities(this.SearchclusterIndex);
		        }
		        if(clusterEntities != null && clusterEntities.size() > 0) 
		        {
		            for(StandardEntity entity : clusterEntities) 
		            {
		            	if(entity instanceof Road)
		            	{	Road road=(Road) entity;
		            		boolean ifadd=true;
		            		for(EntityID entityid:road.getNeighbours())
		            		{
		            			if(this.worldInfo.getEntity(entityid) instanceof Building)
		            			{	
		            				ifadd=false;
		            				break;
		            			}
		            		}
		            		if(ifadd)
		            			this.unsearchedRoadIDs.add(entity.getID());
		            	}
		            }
		        } 
		        else 
		        {
		            this.unsearchedRoadIDs.addAll(this.worldInfo.getEntityIDsOfType(
		                    StandardEntityURN.ROAD,
		                    GAS_STATION,
		                    AMBULANCE_CENTRE,
		                    FIRE_STATION,
		                    POLICE_OFFICE
		            ));
		        }
			}
			else
			{
				this.SearchTime=threshold;
				this.SearchclusterIndex=(this.SearchclusterIndex+1)%this.clustering.getClusterNumber();
				this.unsearchedRoadIDs.clear();
		        Collection<StandardEntity> clusterEntities = null;
		        if(this.clustering != null) {
		            clusterEntities = this.clustering.getClusterEntities(this.SearchclusterIndex);
		        }
		        if(clusterEntities != null && clusterEntities.size() > 0) 
		        {
		            for(StandardEntity entity : clusterEntities) 
		            {
		            	if(entity instanceof Road)
		            	{this.unsearchedRoadIDs.add(entity.getID());}
		            }
		        } 
		        else 
		        {
		            this.unsearchedRoadIDs.addAll(this.worldInfo.getEntityIDsOfType(
		                    StandardEntityURN.ROAD,
		                    GAS_STATION,
		                    AMBULANCE_CENTRE,
		                    FIRE_STATION,
		                    POLICE_OFFICE
		            ));
		        }
			}
		}
		else
		{
			this.unsearchedBuildingIDs.clear();
			int clusterIndex = this.clustering.getClusterIndex(this.agentInfo.getID());
			Collection<StandardEntity> clusterEntities = this.clustering.getClusterEntities(clusterIndex);
			if (clusterEntities != null && clusterEntities.size() > 0) {
				for (StandardEntity entity : clusterEntities) {
					if (entity instanceof Building && entity.getStandardURN() != REFUGE) {
						this.unsearchedBuildingIDs.add(entity.getID());
					}
				}
			} else {
				this.unsearchedBuildingIDs.addAll(this.worldInfo.getEntityIDsOfType(BUILDING, GAS_STATION, AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE));
			}
		}
	}
	private void JudgeFire()
	{
		this.ContainFire=false;
		for(StandardEntity entity:this.clustering.getClusterEntities(this.SearchclusterIndex))
		{
			if(entity instanceof Building && ((Building) entity).isOnFire())
			{
				this.ContainFire=true;
				break;
			}
		}
	}
	private boolean Burnout()
	{
		for(EntityID entityid:this.worldInfo.getChanged().getChangedEntities())
		{
			if(this.worldInfo.getEntity(entityid) instanceof Building)
			{
				Building building=(Building) this.worldInfo.getEntity(entityid);
				if(building.isFierynessDefined()&&building.getFieryness()!=8)
					return false;
			}
		}
		return true;
	}
	@Override
	public EntityID getTarget() {
		return this.result;
	}
 	private void reflectMessage(MessageManager messageManager) {
	        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
	        changedEntities.add(this.agentInfo.getID());
	        int time = this.agentInfo.getTime();

	        for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
	            Class<? extends CommunicationMessage> messageClass = message.getClass();
	            if(messageClass == MessageBuilding.class) {
	                MessageBuilding mb = (MessageBuilding)message;
	                if(!changedEntities.contains(mb.getBuildingID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mb);
	                }
	            } else if(messageClass == MessageRoad.class) {
	                MessageRoad mr = (MessageRoad)message;
	                if(mr.isBlockadeDefined() && !changedEntities.contains(mr.getBlockadeID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mr);
	                }
	                
	            } else if(messageClass == MessageCivilian.class) {
	                MessageCivilian mc = (MessageCivilian) message;
	                if(!changedEntities.contains(mc.getAgentID())){
	                    MessageUtil.reflectMessage(this.worldInfo, mc);
	                }
	                
	            } else if(messageClass == MessageAmbulanceTeam.class) {
	                MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;
	                if(!changedEntities.contains(mat.getAgentID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mat);
	                }
	               
	            } else if(messageClass == MessageFireBrigade.class) {
	                MessageFireBrigade mfb = (MessageFireBrigade) message;
	                if(!changedEntities.contains(mfb.getAgentID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mfb);
	                }
	               
	            } else if(messageClass == MessagePoliceForce.class) {
	                MessagePoliceForce mpf = (MessagePoliceForce) message;
	                if(!changedEntities.contains(mpf.getAgentID())) {
	                    MessageUtil.reflectMessage(this.worldInfo, mpf);
	                }
	                
	            }
	        }
	    }

}
